########################################################################
############## Definition of all wrappers for 1D plotting ##############
########################################################################


####################################
# Generalized lines
####################################
def axline(x=None,y=None,a=None,b=None,
           xlim=None,ylim=None,xinvert=False,yinvert=False,xlog=False,ylog=False,title=None,
           xlabel=None,ylabel=None,label=None,grid=None,ax=None,plot_kw={},**kwargs):
    
    """Generalised axis lines.
    
    This function aims to generalise the usage of axis lines calls (axvline/axhline) together and
    to allow lines to be specified by a slope/intercept according to the function y=a*x + b.
    
    Parameters
    ----------
    x : int or list, optional 
        x position(s) in data coordinates for a vertical line(s).
    y : int or list, optional
        y position(s) in data coordinates for a horizontal line(s).
    a : int or list, optional
        Slope(s) of diagonal axis line(s), defaults to 1 if not specified when b is given.
    b : int or list, optional
        Intercept points(s) of diagonal axis line(s), defaults to 0 if not specified when a is given.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    xinvert : bool or list, optional
        If true inverts the x-axis.
    yinvert : bool or list, optional
        If true inverts the y-axis.
    xlog : bool or list, optional
        If True the scale of the x-axis is logarithmic.
    ylog : bool or list, optional
        If True the scale of the x-axis is logarithmic.
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    label : str, optional
        Sets the legend for the plot.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    ax : pyplot.Axes or list-like, optional
        Specifies the axes on which to plot the lines, defaults to the current axes.
    plot_kw : dict, optional
        Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are Line2D properties.
    **kwargs: Line2D properties, optional
        kwargs are used to specify matplotlib specific properties such as linecolor, linewidth,
        antialiasing, etc. A list of available `Line2D` properties can be found here: 
        https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D
    
    Returns
    -------
    lines
        A list of Line2D objects representing the plotted lines. If more than one subplot and/or line is plotted,
        the resulting list will be multi-dimensional with the dimensions corresponding to the axes/lines.
    
    """
    
    from matplotlib.pyplot import plot, legend, gca, sca
    from splotch.base_func import axes_handler,plot_finalizer,dict_splicer,is_numeric
    from numpy import ndarray, shape, array, squeeze
    
    # Set the current axis
    if ax is not None:
        if isinstance(ax, (list, tuple, ndarray)):
            if len(shape(ax)) > 1: # If ax array is multi-dimensional, flatten it
                ax = array(ax).flatten()
            old_ax=axes_handler(ax[0])
        else:
            ax = [ax] # Axis must be a list to be enumerated over
            old_ax=axes_handler(ax)
    else:
        ax=[gca()]
        old_ax=ax[0]
    
    # Validate input parameters
    if not (any([is_numeric(var) for var in [x,y,a,b]])): # If nothing has been specified
        raise TypeError("axline() missing one of optional arguments: 'x', 'y', 'a' or 'b'")

    for i, val in enumerate([x,y,a,b]):
        if (val is not None):
            try: # Test whether the parameter is iterable
                temp=(k for k in val)
            except TypeError: # If not, convert to a list
                if   (i == 0): x=[x]
                elif (i == 1): y=[y]
                elif (i == 2): a=[a]
                elif (i == 3): b=[b]
    
    if (x is not None and y is not None): # Check whether both x and y were specified
        raise ValueError("'x' and 'y' cannot be both specified")
    
    if (x is not None): # Check conditions if x specified
        if (any([a,b])): # Should not specify a or b, if x given.
            raise ValueError("'{0}' cannot be specified if x specified".format('a' if a else 'b'))
        L=len(x)
    
    if (y is not None): # Check conditions if y specified
        if (any([a,b])): # Should not specify a or b, if y given.
            raise ValueError("'{0}' cannot be specified if y specified".format('a' if a else 'b'))
        L=len(y)
    
    if (a is not None):
        if (b is None): # If no intercept specified
            b=[0]*len(a) # set b to 0 for all a
        else:
            if (len(b) == 1):
                b=[b[0]]*len(a)
            elif (len(b) != len(a)):
                if (len(a) == 1):
                    a=[a[0]]*len(b)
                else:
                    raise ValueError(f"Length of 'a' ({len(a)}) and length of 'b' ({len(b)}) must be equal or otherwise 1")
        L=len(a)
    elif (b is not None):
        if (a is None): # If no slope specified
            a=[1]*len(b) # set a to 1 for all b
        L=len(b)
    
    if type(label) is not list:
        label=[label for i in range(L)]
    elif (len(label) != L):
        raise ValueError("Length of label list ({0}) must match the number of lines given ({1}).".format(len(label),L))
    
    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    #plot_par={**plot_kw, **kwargs} # For Python > 3.5
    plot_par=plot_kw.copy()
    plot_par.update(kwargs)
    
    # Create 'L' number of plot kwarg dictionaries to parse into each plot call
    plot_par=dict_splicer(plot_par,L,[1]*L)
    
    lines=[[] for kk in range(len(ax))] # Initialise list which contains each Line2D object
    for jj, axis in enumerate(ax): # Loop over all axes
        sca(axis)
        if (x is not None):
            for ii, xx in enumerate(x):
                lines[jj].append(axis.axvline(x=xx,**plot_par[ii],label=label[ii]))
        if (y is not None):
            for ii, yy in enumerate(y):
                lines[jj].append(axis.axhline(y=yy,**plot_par[ii],label=label[ii]))
        if (a is not None):
            for ii, (aa,bb) in enumerate(zip(a,b)):
                xLims = axis.get_xlim() if xlim is None else xlim
                yLims = axis.get_ylim() if ylim is None else ylim

                lines[jj].append(axis.plot([xLims[0],xLims[1]],[aa*xLims[0]+bb,aa*xLims[1]+bb],label=label[ii],**plot_par[ii])[0])
                axis.set_xlim(xLims)
                axis.set_ylim(yLims)

        plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)

        # Autoscale the axes if needed
        if xlim is None: axis.autoscale(axis='x')
        if ylim is None: axis.autoscale(axis='y')

    if old_ax is not None: # Reset the previously set axis
        old_ax=axes_handler(old_ax)

    return squeeze(lines).tolist() # Reduce the dimensionality of the lines, if needed


####################################
# Broken axis plot
####################################
def brokenplot(x,y=None,xbreak=None,ybreak=None,xlim=None,ylim=None,sep=0.05,
               xinvert=False,yinvert=False,xlog=False,ylog=False,title=None,
               xlabel=None,ylabel=None,label=None,lab_loc=0,ax=None,grid=None,plot_kw={},**kwargs):

    """Broken Axis Plot Function
    
    Creates a standard plot call with an axis break at `xbreak` or `ybreak` for vertical or
    horizontal breaks.
    
    Parameters
    ----------
    x : array-like or list
        If list it is assumed that each elemement is array-like. If y is not given, the given values
        pass to y and a numpy array is generated with numpy.arange() for the x values.
    y : array-like or list, optional
        If list it is assumed that each elemement is array-like.
    xbreak/ybreak : float or tuple-like, required
        The location(s) of the vertical or horizontal breaks is controlled by xbreak or ybreak, respectively.
        The value can be a single location or a tuple defining the (start, stop) points of the break.
        Only one coordinate can be broken in a given plot.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    sep : float, optional, default: 0.05
        The separation size of the axis break, given as a fraction of the axis dimensions.
    xinvert : bool or list, optional
        If true inverts the x-axis.
    yinvert : bool or list, optional
        If true inverts the y-axis.
    xlog : bool or list, optional
        If True the scale of the x-axis is logarithmic.
    ylog : bool or list, optional
        If True the scale of the x-axis is logarithmic.
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    label : str, optional
        Sets the legend for the plot.
    lab_loc : int, optional
        Defines the position of the legend
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    plot_kw : dict, optional
        Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are Line2D properties.
    **kwargs: Line2D properties, optional
        kwargs are used to specify matplotlib specific properties such as linecolor, linewidth,
        antialiasing, etc. A list of available `Line2D` properties can be found here: 
        https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D
    
    Returns
    -------
    lines
        A list of Line2D objects (paired as tuples) representing the plotted data.
        The lines are given as pairs to correspond to the separate lines either side of the x/ybreak.
    
    """
    from .base_func import axes_handler,dict_splicer,plot_finalizer
    
    from numpy import shape, arange, ndarray
    from matplotlib.pyplot import plot, legend, show, sca, gca
    from matplotlib.transforms import Bbox
    
    if ax is not None:
        old_axes=axes_handler(ax)
    else:
        ax=gca()
        old_axes=ax

    if type(x) is not list or len(shape(x))==1:
        x=[x]
    L=len(x)
    if y is None:
        y=x
        x=[arange(len(x[i])) for i in range(L)]
    else:
        if type(y) is not list or len(shape(y))==1:
            y=[y]
    if type(label) is not list:
        label=[label for i in range(L)]
    
    # Validate x/ybreak
    if (xbreak == None):
        raise ValueError("Require either xbreak/ybreak to be specified.")

    if (ybreak != None):
        raise NotImplementedError("ybreak not yet implemented.")


    if type(xbreak) not in [list,tuple,ndarray]:
        xbreak=(xbreak, xbreak)
    else:
        if (len(xbreak) != 2):
            raise ValueError("xbreak must be a single value of a tuple-like list of two elements.")
    
    
    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    #plot_par={**plot_kw, **kwargs} # For Python > 3.5
    plot_par=plot_kw.copy()
    plot_par.update(kwargs)
    
    # Create 'L' number of plot kwarg dictionaries to parse into each plot call
    plot_par=dict_splicer(plot_par,L,[1]*L)
    
    # Get the original axis position
    pos0=ax.get_position(original=True)
    width0, height0=pos0.x1 - pos0.x0, pos0.y1 - pos0.y0
    
    lines=[] # Initialising list which contains each line
    for i in range(L):
        # First side plot call
        l1=ax.plot(x[i],y[i],label=label[i],**plot_par[i])
        
        # Get the axis limits if not already specified
        xlims=ax.get_xlim() if xlim == None else xlim
        ylims=ax.get_ylim() if ylim == None else ylim
    
        # Define the positions of the two separated axes
        if (i == 0):
            pos1=Bbox(list(pos0.get_points()))
            pos1.x1=pos1.x0 + (pos1.x1-pos1.x0)*(sum(xbreak)/2-xlims[0])/(xlims[1]-xlims[0]) - sep*(pos1.x1-pos1.x0)/2
            
            pos2=Bbox(list(pos0.get_points()))
            pos2.x0=pos2.x0 + (pos2.x1-pos2.x0)*(sum(xbreak)/2-xlims[0])/(xlims[1]-xlims[0]) + sep*(pos2.x1-pos2.x0)/2
            
            ax.set_position(pos1) # Resize the first axis
            ax2=ax.figure.add_axes(pos2) # Add and duplicate the plotting in the second axis
            
            # Set the new axis limits at the break point
            ax.set_xlim(xlims[0],xbreak[0])
            ax2.set_xlim(xbreak[1],xlims[1])
        
        # Second side plot call
        l2=ax2.plot(x[i],y[i],label=None,**plot_par[i])

        lines.append((*l1,*l2)) # Add line as tuple of both axes.
        
        width1, height1=pos1.x1 - pos1.x0, pos1.y1 - pos1.y0
        width2, height2=pos2.x1 - pos2.x0, pos2.y1 - pos2.y0
        
        dx1, dy1=0.01 * width0/(width0-width1-sep/2), height1*0.025
        
        dash_kw=dict(transform=ax2.transAxes, color='black', linestyle='-', marker='', clip_on=False)
        ax2.plot((0 - dx1, 0 + dx1), (0 - dy1, 0 + dy1), **dash_kw)  # bottom-right diagonal
        ax2.plot((0 - dx1, 0 + dx1), (1 - dy1, 1 + dy1), **dash_kw)  # top-right diagonal
        
        dx2, dy2=0.01 * width0/(width0-width2-sep/2), height2*0.025
        dash_kw.update(transform=ax.transAxes)  # switch to the left axes
        ax.plot((1 - dx2, 1 + dx2), (0 - dy2, 0 + dy2), **dash_kw)  # bottom-left sep/5iagonal
        ax.plot((1 - dx2, 1 + dx2), (1 - dy2, 1 + dy2), **dash_kw)  # top-left sep/5iagonal
        
        ax.spines['right'].set_visible(False)
        ax.tick_params(labelright=False,which='both')  # don't put tick labels at the top
        ax.yaxis.tick_left()
        
        ax2.spines['left'].set_visible(False)
        ax2.tick_params(labelleft=False,which='both')  # don't put tick labels at the top
        ax2.yaxis.tick_right()

        # Check that there is no duplicate ticks over both axes    
    if (xbreak):
        if (ax.get_xticks()[-1] == ax2.get_xticks()[0]):
            if (xbreak[0] >= (xlims[0] + xlims[1])*0.5):
                ax.set_xticks(ax.get_xticks()[:-1]) # Remove duplicate tick on left side
            else:
                ax2.set_xticks(ax2.get_xticks()[1:]) # Remove duplicate tick on right side
    sca(ax)
    
    if any(label):
        ax.legend(loc=lab_loc)
    
    plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
    
    if old_axes is not ax:
        old_axes=axes_handler(old_axes)

    return (lines[0] if len(lines) == 1 else lines)

########################################
# Curves from mathematical expressions #
########################################
def curve(expr, var=None, subs={}, orientation='horizontal', permute=False, bounds=None, num=101, 
          xlim=None, ylim=None, xinvert=False, yinvert=False, xlog=False, ylog=False, grid=None, 
          title=None, xlabel=None, ylabel=None, label=True, uselatex=True, ax=None, plot_kw={}, **kwargs):
    """Plot Mathematical Expressions
    
    Plot the curve(s) corresponding to mathematical expressions over a given range across the independent variable.
    Expressions can be given with multiple variables, whereby one is taken to be the independent variable and all 
    others are substitution variables, which can take a variable number of values producing the corresponding number
    of curves.
    
    Parameters
    ----------
    expr : str, sympy.Expr, callable() or list-like
        An expression parsed either as a string, sympy expression or callable (function or lambda)
        which will be evaluated by the function in the range of `bounds`.
    var : str or sympy symbol, default: 'x'
        The independent variable on which to evaluate the expression (i.e. the variable on the x-axis).
        This defaults to the first non-numeric element of the expression or otherwise simply assumes
        this to be 'x' (if `orientation='horizotnal'`, else 'y').
    subs : dict, optional
        If `expr` contains more symbols than the independent variable `var`, this dictionary will
        substitute numerical values for all additonal symbols given. `subs` is required if additional
        symbols are specified in the expression.
    permute : bool, optional (default: False)
        If `subs` contains more than one substitution variable with multiple values, setting `permute=True`
        will generate curves for each permutation of values given.
        For example, `splotch.curve('a*x^2 + b',subs = {'a'=[1,2],'b'=[3,4,5]})` will create six curves.
    orientation : str, optional (default: 'horizontal')
        The orientation of the independent axis, i.e. whether the independent variable is defined along
        the x-axis ('horizontal') or the y-axis ('vertical') of the plot.
        splotch.curve('a*x') is notationally the same as splotch.curve('1/a*y',var='y',orientation='vertical'). 
    bounds : list-like, optional
        The range over which the function will be plotted. If not given, these default to
        the current bounds of the plot.
    num : int, optional (default: 101)
        The number of values along the independent variable on which to evaulate `expr`.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    xinvert : bool or list, optional
        If true inverts the x-axis.
    yinvert : bool or list, optional
        If true inverts the y-axis.
    xlog : bool or list, optional
        If True the scale of the x-axis is logarithmic.
    ylog : bool or list, optional
        If True the scale of the x-axis is logarithmic.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    label : bool, str or list-like, optional (default: True)
        Sets the label(s) of the curve(s) for the legend. `label` can be one of:
            - `True`:
                Creates a label for every curve defined by `subs`. The label will list all values
                `subs` that produced the curve. If a parameter in `subs` has only one value
                (i.e. constant amongst all curves), it will not appear in the label.
            - `str`:
                If a single string is given, only one label will be shown and all curves will
                be shown as the label handle.
            - list-like (length must equal to number of curves):
                A label given to each curve that is produced.
            - `False` or `None`:
                No label will be assigned and a legend will not be shown.
    uselatex : bool, optional (default: True)
        If `label==True`, sets whether to use LaTeX when creating the labels for legend.
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    plot_kw : dict, optional
        Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are Line2D
        properties. It is recommended that kwargs be parsed implicitly through **kwargs
        for readability.
    **kwargs: Line2D properties, optional
        kwargs are used to specify matplotlib specific properties such as linecolor, linewidth, 
        antialiasing, etc. A list of available `Line2D` properties can be found here: 
        https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D
    
    Returns
    -------
    curves : list of (or single) pyplot.Line2D object(s)
        A list of Line2D objects created for each curved create by `subs`.
    expr : Sympy.Expr
        If expr was given as a string, this returns the sympy expression created from `sympy.sympify()`.
        Otherwise, simply returns the `expr` that was given.
    
    """
    
    from splotch.base_func import axes_handler,dict_splicer,plot_finalizer,simpler_dict_splicer
    
    from sympy import symbols, sympify, Expr, latex
    from sympy.utilities.lambdify import lambdify
    from numpy import linspace, logspace, log10, empty, array, full_like, meshgrid, prod
    from collections import Iterable
    
    from matplotlib.pyplot import plot, legend, gca
    from matplotlib.legend_handler import HandlerPathCollection, HandlerLine2D, HandlerTuple
    from matplotlib import rcParams
    from matplotlib.legend import Legend
    
    if ax is not None:
        old_axes=axes_handler(ax)
    else:
        ax=gca()
        old_axes=ax
        
    # Assign bounds if none given  
    if (bounds == None):
        if orientation == 'horizontal':
            bounds = xlim if xlim is not None else ax.get_xlim()
        else:
            bounds = ylim if ylim is not None else ax.get_ylim()
    
    # Parse expression
    isfunc=False
    if (isinstance(expr, str)):
        expr=sympify(expr)
    elif (callable(expr)):
        isfunc=True
    elif (isinstance(expr, Expr)):
        pass
    else:
        raise TypeError(f"`expr` must be of type `str`, sympy.Expr or callable, instead got {type(expr)}.")
    
    
    if not isfunc: # expr is a Sympy expression
        symbols=expr.free_symbols # Get all of the Symbols in the expression
        symbolkeys=[str(symb) for symb in symbols] # convert these to strings, instead of sympy.Symbols
        
        if var is None: # Assume independent variable is 'x', otherwise, assume the first symbol.
            if orientation == 'horizontal':
                var = 'x' #if 'x' in symbolkeys or len(symbolkeys)==0 else None
            else: # first test for 'y' as an independent variable, then default to x.
                if 'y' in symbolkeys:
                    var = 'y'
                else:
                    var = 'x' #if 'x' in symbolkeys or len(symbolkeys)==0 else symbolkeys[0]
        
        # Validate the substitution variable names
        if subs is None: subs = dict()
        if (var in list(subs)):
            raise ValueError(f"Independent variable '{var}' should not be in subs")
            
        for key in list(subs): 
            if (key not in symbolkeys):
                raise KeyError(f"Substitution variable '{key}' does not exist in 'expr'")
        
        # Check for missing substitution variables
        missing = set(symbolkeys) - set(subs) - set(var)
        if len(missing) > 0:
            raise TypeError(f"`expr` missing {len(missing)} required substitution variable{'s' if len(missing) > 1 else ''}: {list(missing)}")
    
    # The lengths of each substitute value list, len=1 if just a single value
    lens=[len(subs[key]) if (isinstance(subs[key], Iterable) and type(subs[key])!=str) else 1 for key in list(subs)]
    if (permute == True):
        L = prod(lens)
        perms=array(meshgrid(*subs.values())).reshape(len(subs),-1)
        permsubs={}
        for ii, key in enumerate(list(subs)):
            permsubs[key]=perms[ii]
        subsarr=simpler_dict_splicer(permsubs,L,[1]*L)
    else:
        L=max(lens) if len(lens) > 0 else 1
        subsarr=simpler_dict_splicer(subs,L,[1]*L)
    
    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    plot_par={**plot_kw, **kwargs}
    
    # Create 'L' number of plot kwarg dictionaries to parse into each plot call
    plot_par=dict_splicer(plot_par,L,[1]*L)

    # Create the legend object
    if bool(label) == False: # label was `None` or `False`
        labellist = [None]*L
    elif label == True: # Auto-generate labels
        if subsarr == [{}]:
            labellist = [f"${latex(expr)}$" if uselatex else str(expr)]
        else:
            labellist = []
            exprstr = f"${latex(expr)}$" if uselatex else str(expr)
            for ii in range(L): # Make a label for each of sub values
                if uselatex:
                    labellist.append(f"{exprstr} (" + "; ".join( [f"${key}$={subsarr[ii][key]}" for jj, key in enumerate(list(subsarr[ii])) ] ) +")" ) # join substitute strings together
                else:
                    labellist.append(f"{exprstr} (" + "; ".join( [f"{key}={subsarr[ii][key]}" for jj, key in enumerate(list(subsarr[ii])) ] ) +")" ) # join substitute strings together
    elif isinstance(label,str): # A single string
        labellist = [label]*L
    else:
        try: # Test whether the parameter is iterable
            _ = (k for k in label)
            if (len(label) != L):
                raise TypeError(f"Number of labels ({len(label)}) does not match the number of curves ({L}).")
            else:
                labellist = label
        except TypeError: # was not an iterable
            raise TypeError(f"`label` of type {type(label)} is not recognised.")
                                 
    # Create and plot the curves
    vararr=logspace(*log10(bounds),num=num) if xlog else linspace(*bounds,num=num)
    curves=[None]*L
    for ii in range(L):
        if (isfunc):
            curvearr=expr(vararr, **subsarr[ii])
        else:
            lamb = lambdify(var, expr.subs(subsarr[ii]), modules='numpy') # returns a numpy-ready function
            
            if expr.subs(subsarr[ii]).is_constant():
                func = lambda x: full_like(x, lamb(x))
            else:
                func = lambda x: lamb(array(x))
            
            curvearr = func(vararr)
            
        curves[ii]=plot(vararr if orientation=='horizontal' else curvearr,
                        curvearr if orientation=='horizontal' else vararr,
                        label=labellist[ii],**plot_par[ii])[0]
    
    plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
    
    # Autoscale the axes if needed
    if xlim is None: ax.autoscale(axis='x')
    if ylim is None: ax.autoscale(axis='y')
    
    if old_axes is not ax:
        old_axes=axes_handler(old_axes)
    
    return(curves[0] if len(curves)==1 else curves, expr)


##################################################
# Piecewise curves from mathematical expressions #
##################################################
def curve_piecewise(expr, var=None, subs={}, orientation='horizontal', bounds=None, intervals=[], permute=False, num=101, 
                   xlim=None, ylim=None, xinvert=False, yinvert=False, xlog=False, ylog=False, grid=None, 
                   title=None, xlabel=None, ylabel=None, label=True, uselatex=True, ax=None, plot_kw={}, **kwargs):
    """Plot Piecewise Mathematical Expressions
    
    Plot the curve(s) corresponding to mathematical expressions over a given range across the independent variable.
    Expressions can be given with multiple variables, whereby one is taken to be the independent variable and all 
    others are substitution variables. This function can only accept one value per substitution variable.
    
    Parameters
    ----------
    expr : str, sympy.Expr, callable() or list-like
        An expression parsed either as a string, sympy expression or callable (function or lambda)
        which will be evaluated by the function in the range of `bounds`. A piece-wise function is defined 
        by parsing a list of expressions. The piece-wise functionality must also be reflected in `bounds`.
    var : str or sympy symbol, required.
        The independent variable on which to evaluate the expression (i.e. the variable on the x-axis).
    subs : dict, optional
        If `expr` contains more symbols than the independent variable `var`, this dictionary will
        substitute numerical values for all additonal symbols given. `subs` is required if additional
        symbols are specified in the expression.
    orientation : str, optional (default: 'horizontal')
        The orientation of the independent axis, i.e. whether the independent variable is defined along
        the x-axis ('horizontal') or the y-axis ('vertical') of the plot.
        splotch.curve('a*x') is notationally the same as splotch.curve('1/a*y',var='y',orientation='vertical'). 
    bounds : list-like, optional
        The range over which the function will be plotted. If not given, these default to the current bounds 
        of the axes being plotted onto, given by `ax`.
    intervals : list-like, required
        The points at which the curve partitions each of the N expressions given in `expr`. The values must lie
        lie within `bounds` as these are assumed to be the minimum and maximum points of the intervals.
    num : int, optional (default: 101)
        The number of values along the independent variable on which to evaulate `expr`.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    xinvert : bool or list, optional
        If true inverts the x-axis.
    yinvert : bool or list, optional
        If true inverts the y-axis.
    xlog : bool or list, optional
        If True the scale of the x-axis is logarithmic.
    ylog : bool or list, optional
        If True the scale of the x-axis is logarithmic.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    label : bool, str or list-like, optional (default: True)
        Sets the label(s) of the curve(s) for the legend. `label` can be one of:
            - `True`:
                Creates a label for every curve defined by `subs`. The label will list all values
                `subs` that produced the curve. If a parameter in `subs` has only one value
                (i.e. constant amongst all curves), it will not appear in the label.
            - `str`:
                If a single string is given, only one label will be shown and all curves will
                be shown as the label handle.
            - list-like (length must equal to number of curves):
                A label given to each curve that is produced.
            - `False` or `None`:
                No label will be assigned and a legend will not be shown.
    uselatex : bool, optional (default: True)
        If `label==True`, sets whether to use LaTeX when creating the labels for legend.
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    plot_kw : dict, optional
        Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are Line2D
        properties. It is recommended that kwargs be parsed implicitly through **kwargs
        for readability.
    **kwargs: Line2D properties, optional
        kwargs are used to specify matplotlib specific properties such as linecolor, linewidth, 
        antialiasing, etc. A list of available `Line2D` properties can be found here: 
        https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D
    
    Returns
    -------
    curves : list of (or single) pyplot.Line2D object(s)
        A list of Line2D objects created for each curved create by `subs`.
    expr : Sympy.Expr
        If expr was given as a string, this returns the sympy expression created from `sympy.sympify()`.
        Otherwise, simply returns the `expr` that was given.
    
    """
    
    from splotch.base_func import axes_handler,dict_splicer,plot_finalizer,simpler_dict_splicer
    
    from sympy import symbols, sympify, Expr, latex
    from sympy.utilities.lambdify import lambdify
    from numpy import linspace, logspace, log10, empty, array, full_like, meshgrid, prod, piecewise
    from collections import Iterable
    
    from matplotlib.pyplot import plot, legend, gca
    from matplotlib.legend_handler import HandlerPathCollection, HandlerLine2D, HandlerTuple
    from matplotlib import rcParams
    from matplotlib.legend import Legend
    from matplotlib.collections import LineCollection
    
    if ax is not None:
        old_axes=axes_handler(ax)
    else:
        ax=gca()
        old_axes=ax
        
    # Assign bounds if none given  
    if (bounds == None):
        if orientation == 'horizontal':
            bounds = xlim if xlim is not None else ax.get_xlim()
        else:
            bounds = ylim if ylim is not None else ax.get_ylim()

    # Check if iterable
    try: # duck-type check
        _ = (k for k in expr)
        if isinstance(expr,str):
            expr = [expr]
        elif isinstance(expr, tuple):
            expr = list(expr)
    except (TypeError):
        expr = [expr]
    
    # Parse expressions
    isfunc=[False]*len(expr)
    for ii in range(len(expr)):
        if (isinstance(expr[ii], str)):
            expr[ii]=sympify(expr[ii])
        elif (callable(expr[ii])):
            isfunc[ii] = True
        elif (isinstance(expr[ii], Expr)):
            pass
        else:
            raise TypeError(f"Elements of `expr` must be of type `str`, sympy.Expr or callable, instead got {type(expr[ii])}.")
            
    if not (all(isfunc) or all([not b for b in isfunc])): # Must either be all expressions or all callables
        raise TypeError("`expr` cannot mix callable functions with expressions.")
        
    # Validate intervals is given in the correct format and the correct number given within the bounds.
    try:
        _ = (k for k in intervals)
        
        if isinstance(intervals, tuple):
            intervals = list(intervals)
        if len(intervals) == 0: # if no intervals are given, set the interval to be the ending bound as a placeholder
            if len(expr) == 1:
                intervals = [bounds[-1]]
            else:
                raise ValueError(f"No intervals given for {len(expr)} expressions.")
        elif len(intervals) != len(expr) - 1: 
            raise ValueError(f"There should be N-1 intervals for N expressions, instead received {len(intervals)} intervals for {len(expr)} expressions.")
        else: # If intervals are given, ensure they are within the bounds given.
            if min(intervals) <= bounds[0]:
                raise ValueError(f"The minimum interval value should be within the current bounds ({bounds[0]}, {bounds[1]}).")
            elif max(intervals) >= bounds[1]:
                raise ValueError(f"The maximum interval value should be within the current bounds ({bounds[0]}, {bounds[1]}).")
    except (TypeError):
        intervals = [intervals]
        

    # Validate the substitution variable names
    symbolkeysArr = []
    if (var in list(subs)):
        raise ValueError(f"Independent variable '{var}' should not be in subs")
        
    if subs is None: subs = dict()
    if not any(isfunc): # expr contains Sympy expressions
        symbols = expr[0].free_symbols
        for ii in range(len(expr)):
            symbols=expr[ii].free_symbols # Get expression Symbols
            symbolkeys=[str(symb) for symb in symbols] # convert these to strings, instead of sympy.Symbols
            
            if var is None: # Assume independent variable is 'x', otherwise, assume the first symbol.
                if orientation == 'horizontal':
                    var = 'x' #if 'x' in symbolkeys or len(symbolkeys)==0 else None
                else: # first test for 'y' as an independent variable, then default to x.
                    if 'y' in symbolkeys:
                        var = 'y'
                    else:
                        var = 'x' #if 'x' in symbolkeys or len(symbolkeys)==0 else symbolkeys[0]
                        
            # remove the independent variable from the symbols and append to list
            symbolkeysArr += symbolkeys
        
        # Check for missing substitution variables
        missing = set(symbolkeysArr) - set(subs) - set(var)
        if len(missing) > 0:
            raise TypeError(f"`expr` missing {len(missing)} required substitution variable{'s' if len(missing) > 1 else ''}: {list(missing)}")

        # Check for any substitution variables that are not in any expressions
        for key in list(subs): 
            if not any([key in symb for symb in symbolkeysArr]):
                raise KeyError(f"Substitution variable '{key}' does not exist in any 'expr'")

    # The lengths of each substitute value list, len=1 if just a single value
    lens = [len(subs[key]) if (isinstance(subs[key], Iterable) and type(subs[key])!=str) else 1 for key in list(subs)]
    if (permute == True):
        L = prod(lens)
        perms=array(meshgrid(*subs.values())).reshape(len(subs),-1)
        permsubs={}
        for ii, key in enumerate(list(subs)):
            permsubs[key]=perms[ii]
        subsarr=simpler_dict_splicer(permsubs,L,[1]*L)
    else:
        L=max(lens) if len(lens) > 0 else 1
        subsarr=simpler_dict_splicer(subs,L,[1]*L)
    
    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    plot_par={**plot_kw, **kwargs}
    
    # Create 'L' number of plot kwarg dictionaries to parse into each plot call
    plot_par=dict_splicer(plot_par,L,[1]*L)

    # Create the legend object
    if bool(label) == False: # label was `None` or `False`
        labellist = None
    
    elif label == True: # Auto-generate labels
        if subsarr == [{}]:
            labellist = [f"${latex(expr)}$" if uselatex else str(expr)]
        else:
            labellist = []
            exprstr = f"${latex(expr)}$" if uselatex else str(expr)
            for ii in range(L): # Make a label for each of sub values
                if uselatex:
                    labellist.append(f"{exprstr} (" + "; ".join( [f"${key}$={subsarr[ii][key]}" for jj, key in enumerate(list(subsarr[ii])) ] ) +")" ) # join substitute strings together
                else:
                    labellist.append(f"{exprstr} (" + "; ".join( [f"{key}={subsarr[ii][key]}" for jj, key in enumerate(list(subsarr[ii])) ] ) +")" ) # join substitute strings together
    
    elif isinstance(label,str): # A single string
        labellist = [label]*L
    
    else:
        try: # Test whether the parameter is iterable
            _ = (k for k in label)
        except TypeError: # was not an iterable
            raise TypeError(f"`label` of type {type(label)} is not recognised.")
            
        if (len(label) != L):
            raise TypeError(f"Number of labels ({len(label)}) does not match the number of curves ({L}).")
        else:
            labellist = label

    
    vararr = logspace(*log10(bounds),num=num) if xlog else linspace(*bounds,num=num)
    intervals.append(vararr[-1])
    curves=[None]*L
    for ii in range(L): 
        lines = []
        start = 0
        for jj in range(len(expr)):
            end = array(abs(vararr - intervals[jj])).argmin()
            if any(isfunc):
                lines.append( list( zip(vararr[start:end+1],expr[jj](vararr[start:end+1],**subsarr[ii])) ) )
            else:
                lamb = lambdify(var, expr[jj].subs(subsarr[ii]), modules='numpy')
                if expr[jj].subs(subsarr[ii]).is_constant():
                    func = lambda x: full_like(x, lamb(x))
                else:
                    func = lambda x: lamb(array(x))
                
                if orientation == 'horizontal':
                    lines.append( list( zip(vararr[start:end+1],func(vararr[start:end+1])) ) )
                else:
                    lines.append( list( zip(func(vararr[start:end+1]),vararr[start:end+1]) ) )
                    
            start = end # move to next interval

        curves[ii] = LineCollection(lines, label=labellist[ii], **plot_par[ii])
        ax.add_collection(curves[ii])

    plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
    
    if xlim is None: ax.autoscale(axis='x')
    if ylim is None: ax.autoscale(axis='y')
    
    if old_axes is not ax:
        old_axes=axes_handler(old_axes)
    
    return(curves[0] if len(curves)==1 else curves)


####################################
# 1D histogram and binned statistics
####################################
def hist(data,bin_type=None,bins=None,dens=True,cumul=None,scale=None,weights=None,hist_type=None,v=None,vstat=None,
            xlim=None,ylim=None,nmin=0,xinvert=False,yinvert=False,xlog=False,ylog=None,title=None,xlabel=None,ylabel=None,
            label=None,lab_loc=0,ax=None,grid=None,plot_kw={},output=None,**kwargs):
    
    """1D histogram function.
    
    The plotting is done with pyplot.plot(), so histograms are shown with interpolated curves instead
    of the more common stepwise curve. For this reason splotch.histstep is a better choice for
    small datasets. 
    
    Parameters
    ----------
    data : array-like or list
        If list it is assumed that each elemement is array-like.
    bin_type : {'number','width','edges','equal'}, optional
        Defines how is understood the value given in bins: 'number' for the desired number of bins,
        'width' for the width of the bins, 'edges' for the edges of bins, and 'equal' for making
        bins with equal number of elements (or as close as possible). If not given it is inferred
        from the data type of bins: 'number' if int, 'width' if float and 'edges'if ndarray.
    bins : int, float, array-like or list, optional
        Gives the values for the bins, according to bin_type.
    dens :  bool or list, optional
        If false the histogram returns raw counts.
    cumul : bool or list, optional
        If true, produces a cumulative distribution instead of a histogram.
    scale : float or list, optional
        Scaling to be applied to the counts.
    weights : array-like or None, optional
        An array of weights with the same shape as data. For each value in data, it will only
        contribute its given weight towards the bin count (instead of 1). If dens is True, weights
        will also be normalised so that the integral over the density remains 1. Default: None
    hist_type : str, optional.
        Defines the type of histogram to be drawn. 'line' and 'step' produce lines, with the former
        drawing lines conecting the values of each bin positioned on their centre, and the latter
        drawing a stepwise line, with the edges of each step coinciding with the bin edges.
        'bar' produces a bar plot. All have filled version (i.e., 'linefilled'), which fills the
        space between the edges of the histogram and 0.
    v : array-like or list, optional
        If a valid argument is given in vstat, defines the value used for the binned statistics.
    vstat : str, function  or list, optional
        Must be or contain one of the valid str arguments for the statistics variable
        in scipy.stats.binned_statistic ('mean’, 'median’, 'count’, 'sum’, 'min’ or 'max’) or
        function(s) that takes a 1D array and outputs an integer or float.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    nmin : int, optional (default: 0)
        The minimum number of points required in a bin in order to be plotted.
    xinvert : bool or list, optional
        If true inverts the x-axis.
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    yinvert : bool or list, optional
        If true inverts the y-axis.
    xlog : bool or list, optional
        If True the scale of the x-axis is logarithmic.
    ylog : bool or list, optional
        If True the scale of the x-axis is logarithmic.
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    label : str, optional
        Sets the label for the plot.
    lab_loc : int, optional
        Defines the position of the legend
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    plot_par : dict, optional
        Passes the given dictionary as a kwarg to the plotting function.
    output : boolean, optional
        If True, returns the edges and values of the histogram.
    
    Returns
    -------
    n : list
        List containing the arrays with the values for each histogram drawn. Only provided
        if output is True.
    bins_edges : list
        List containing the arrays with the bin edges for each of the histograms drawn.
        Only provided if output is True.
    """
    
    from numpy import cumsum as np_cumsum, sum as np_sum, max as np_max, min as np_min
    from scipy.stats import binned_statistic
    from numpy import array, ndarray, diff, dtype, histogram, inf, nan, nanmax, nanmean, nanstd, ones, where, shape
    from matplotlib.pyplot import bar, fill_between, gca, legend, plot, rcParams, step
    from .base_func import axes_handler,bin_axis,dict_splicer,plot_finalizer,step_filler
    
    if ax is not None:
        old_axes=axes_handler(ax)
    else:
        ax=gca()
        old_axes=ax
    
    if type(data) not in [list, tuple, ndarray] or (len(shape(data))==1 and array(data).dtype is not dtype('O')):
        data=[data]
    L=len(data)
    if type(bin_type) not in [list, tuple]:
        bin_type=[bin_type]*L
    if type(bins) not in [list, tuple, ndarray] or (len(shape(bin_type))==1):
        if bins is not None:
            bins=[bins]*L
        else:
            bins=[int((len(d))**0.4) for d in data]
    if type(weights) not in [list, tuple, ndarray] or (len(shape(weights))==1):
        weights=[weights]*L
    if type(dens) not in [list, tuple]:
        dens=[dens]*L
    if type(cumul) not in [list, tuple]:
        cumul=[cumul]*L
    if type(scale) not in [list, tuple, ndarray]:
        scale=[scale]*L
    if type(v) not in [list, tuple, ndarray] or (len(shape(v)) == 1):
        v=[v]*L
    if type(nmin) not in [list, tuple, ndarray] or (len(shape(nmin)) == 1):
        nmin=[nmin]*L
    if type(vstat) not in [list, tuple]:
        vstat=[vstat]*L
    if type(label) not in [list, tuple]:
        label=[label for i in range(L)]
    if type(xlim) in [int,float]:
        xlim=[nanmean(data)-xlim*nanstd(data),nanmean(data)+xlim*nanstd(data)]
    
    if None in [ylog,hist_type,output]:
        from .defaults import Params
        if ylog is None:
            ylog=Params.hist1D_yaxis_log
        if hist_type is None:
            hist_type=Params.hist1D_histtype
        if output is None:
            output=Params.hist1D_output
    if type(hist_type) not in [list, tuple, ndarray]:
        hist_type=[hist_type]*L
    
    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    #plot_par={**plot_kw, **kwargs} # For Python > 3.5
    plot_par=plot_kw.copy()
    plot_par.update(kwargs)
    # Check if width is given as a kwarg
    if 'width' in plot_par.keys():
        import warnings
        warnings.warn('Received kwarg width, this will be ignored in the histogram',UserWarning)
        if hist_type!='bar':
            temp=plot_par.pop('width')
    
    # Create 'L' number of plot kwarg dictionaries to parse into each plot call
    plot_par=dict_splicer(plot_par,L,[1]*L)
    
    plot_type={'line':plot,'linefilled':fill_between,'step':step,'stepfilled':step_filler,'bar':bar,'barfilled':bar}
    hist_centre={'line':True,'linefilled':True,'step':False,'stepfilled':False,'bar':False,'barfilled':False}
    bin_edges=[]
    n_return=[]
    
    for i in range(L):
        temp_data,bins_hist,bins_plot=bin_axis(data[i],bin_type[i],bins[i],log=xlog,plot_centre=hist_centre[hist_type[i]])
        n_check=histogram(temp_data,bins=bins_hist,density=False)[0]
        n_check=n_check>=nmin[i]
        if vstat[i]:
            temp_y=binned_statistic(temp_data,v[i],statistic=vstat[i],bins=bins_hist)[0]
        else:
            if scale:
                temp_y=histogram(temp_data,bins=bins_hist,density=False,weights=weights[i])[0]
            else:
                temp_y=histogram(temp_data,bins=bins_hist,density=dens[i],weights=weights[i])[0]
        if cumul[i]:
            temp_y=np_cumsum(temp_y)
            if dens[i]:
                temp_y=temp_y.astype('float')/nanmax(temp_y)
        if scale[i]:
            temp_y=temp_y.astype('float')/scale[i]
            if dens[i]:
                temp_y/=bins_hist[1:]-bins_hist[:-1]
        if ylog:
            temp_y=where(temp_y==0,nan,temp_y)
        temp_y=where(n_check,temp_y,nan)
        y=temp_y
        if hist_type[i]=='step':
            if (ylog or v is not None):
                y=array([y[0]]+[j for j in y])
            else:
                bins_plot=array([bins_plot[0]]+[b for b in bins_plot]+[bins_plot[-1]])
                y=array([0,y[0]]+[j for j in y]+[0])
        if 'bar' in hist_type[i]:
            #prop_cycle=rcParams['axes.prop_cycle']
            #barcolor=prop_cycle.by_key()['color']
            plot_par[i]['width']=diff(bins_plot)
            bins_plot=(bins_plot[0:-1]+bins_plot[1:])/2
            if hist_type[i]=='bar':
                if 'edgecolor' not in plot_par[i].keys():
                    p=plot(bins_plot[0],0)
                    plot_par[i]['edgecolor']=p[0].get_color()
                    p.pop()
                    temp_ax=gca()
                    temp_ax.relim()
                    temp_ax.autoscale()
                plot_par[i]['fill']=False
        plot_type[hist_type[i]](bins_plot,y,label=label[i],**plot_par[i])
        bin_edges.append(bins_plot)
        n_return.append(temp_y)
    if any(label):
        legend(loc=lab_loc)
    
    if ylim == None: # Adjust ylims if None given.
        if not ylog and all([val is None for val in v]): # These automatic limits do not apply when ylog=True or statistics are used.
            ylim = [0, max(np_max(y)*(1+rcParams['axes.ymargin']), gca().get_ylim()[1])]
    
    plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
    if old_axes is not ax:
        old_axes=axes_handler(old_axes)
    if len(n_return)==1:
        n_return=n_return[0]
    if len(bin_edges)==1:
        bin_edges=bin_edges[0]
    if output:
        return(n_return,bin_edges)

####################################
# Standard plots
####################################
def plot(x,y=None,xlim=None,ylim=None,xinvert=False,yinvert=False,xlog=False,ylog=False,title=None,xlabel=None,
            ylabel=None,label=None,lab_loc=0,ax=None,grid=None,plot_kw={},**kwargs):
    
    """Base plotting function.
    
    This is a wrapper for pyplot.plot().
    
    Parameters
    ----------
    x : array-like or list
        If list it is assumed that each elemement is array-like. If y is not given, the given
        values pass to y and a numpy array is generated with numpy.arange() for the x values.
    y : array-like or list, optional
        If list it is assumed that each elemement is array-like.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    xinvert : bool or list, optional
        If true inverts the x-axis.
    yinvert : bool or list, optional
        If true inverts the y-axis.
    xlog : bool or list, optional
        If True the scale of the x-axis is logarithmic.
    ylog : bool or list, optional
        If True the scale of the x-axis is logarithmic.
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    label : str, optional
        Sets the legend for the plot.
    lab_loc : int, optional
        Defines the position of the legend
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    plot_kw : dict, optional
        Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are
        Line2D properties.
    **kwargs: Line2D properties, optional
        kwargs are used to specify matplotlib specific properties such as linecolor, linewidth,
        antialiasing, etc. A list of available `Line2D` properties can be found here: 
        https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D
    
    Returns
    -------
    lines
        A list of Line2D objects representing the plotted data.
    """
    
    from numpy import shape, arange
    from matplotlib.pyplot import gca, plot, legend
    from .base_func import axes_handler,dict_splicer,plot_finalizer
    
    if ax is not None:
        old_axes = axes_handler(ax) # Set current axis to ax and return the previous axis to old_axes.
    else:
        ax=gca()
        old_axes=ax

    if type(x) is not list or len(shape(x))==1:
        x=[x]
    L=len(x)
    if y is None:
        y=x
        x=[arange(len(x[i])) for i in range(L)]
    else:
        if type(y) is not list or len(shape(y))==1:
            y=[y]
    if type(label) is not list:
        label=[label for i in range(L)]
    
    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    #plot_par={**plot_kw, **kwargs} # For Python > 3.5
    plot_par=plot_kw.copy()
    plot_par.update(kwargs)
    
    # Create 'L' number of plot kwarg dictionaries to parse into each plot call
    plot_par=dict_splicer(plot_par,L,[1]*L)
    
    lines=[] # Initialising list which contains each line
    for i in range(L):
        lines += plot(x[i],y[i],label=label[i],**plot_par[i])
    if any(label):
        legend(loc=lab_loc)
    plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
    
    if old_axes is not ax:
        old_axes = axes_handler(old_axes)
    
    return (lines[0] if len(lines) == 1 else lines)




########################################################################
############## Definition of all wrappers for 2D plotting ##############
########################################################################

####################################
# Level contours
####################################
def contour(z, x=None, y=None, filled=None, xlim=None, ylim=None, xinvert=False, yinvert=False, xlog=False, ylog=False, title=None, xlabel=None,
            ylabel=None, lab_loc=0, ax=None, grid=None, plot_kw={}, **kwargs):
    """Level contour plotting function.

    This is a wrapper for pyplot.contour() and pyplot.contourf().

    Parameters
    ----------
    z : array-like
        The height values to draw the contours.
    x : array-like, optional
        Position of data points in the x axis.
    y : array-like, optional
        Position of data points in the y axis.
    filled: boolean, optional
        If True, draws filled contours. If not given defaults to the value defined in splotch.Params.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    xinvert : bool, optional
        If True, inverts the x-axis.
    yinvert : bool, optional
        If True, inverts the y-axis.
    xlog : bool, optional
        If True, the scale of the x-axis is logarithmic. If not given defaults to the value defined in splotch.Params.
    ylog : bool, optional
        If True, the scale of the x-axis is logarithmic. If not given defaults to the value defined in splotch.Params.
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    lab_loc : int, optional
        Defines the position of the legend
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    output : boolean, optional
        If True, returns the edges and values of the underlying histogram plus the levels of the contours.
    plot_kw : dict, optional
        Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are QuadContourSet properties.
    **kwargs: QuadContourSet properties, optional
        kwargs are used to specify matplotlib specific properties such as cmap, linewidths, hatches, etc.
        The list of available properties can be found here:
        https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.contour.html

    Returns
    -------
    bin_edges_x : array
        The bin edges for the x axis.
    bin_edges_y : array
        The bin edges for the y axis.
    n : array
        The values of the underlying histogram.
    l : array
        The levels for the contours.
    """

    from matplotlib.pyplot import gca, contour, contourf
    from .base_func import axes_handler, plot_finalizer

    # Set the current axis
    if ax is not None:
        old_axes = axes_handler(ax)
    else:
        ax = gca()
        old_axes = ax

    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    # plot_par={**plot_kw, **kwargs} # For Python > 3.5
    plot_par = plot_kw.copy()
    plot_par.update(kwargs)

    # Create 'L' number of plot kwarg dictionaries to parse into each plot call
    if filled:
        contourset = contourf(x, y, z, **plot_par)
    else:
        contourset = contour(x, y, z, **plot_par)

    plot_finalizer(xlog, ylog, xlim, ylim, title, xlabel, ylabel, xinvert, yinvert, grid)
    if old_axes is not ax:
        old_axes = axes_handler(old_axes)

    return contourset


####################################
# Contours from density histograms
####################################
def contourp(x, y, percent=None, filled=None, bin_type=None, bins=None, smooth=0.0, max_spacing=True, xlim=None, ylim=None, xinvert=False, yinvert=False,
             xlog=False, ylog=False, title=None, labels=None, xlabel=None, ylabel=None, lab_loc=0, ax=None, grid=None, output=None, plot_kw={}, **kwargs):
    """Contour function, encircling the highest density regions that contain the given percentages of the sample.

    Parameters
    ----------
    x : array-like
        Position of data points in the x axis.
    y : array-like
        Position of data points in the y axis.
    percent : float or array-like, optional.
        The percentages of the sample for the contours to encircle.
    bin_type : {'number', 'width', 'edges', 'equal'}, optional
        Defines the format for the value(s) given in `bins`.
            * 'number' specifies the desired number of bins.
            * 'width' specifies the width of the bins.
            * 'edges' for the edges of bins
            * 'equal' results in bins with an equal number of elements (or as close as possible).
        If `bin_type` not given, it is inferred from the data type of bins, i.e., 'number' if int,
        'width' if float and 'edges' if a list-like object.
    bins : int, float, array-like, optional
        Specifies the value(s) for the bins, according to bin_type.
    smooth : float, optional
        The standard deviation for the Gaussian kernel. Default: 0.0 (No smoothing).
    max_spacing : boolean, optional
        If True, maximises the separation between colours drawn from the colour map. Default: True.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    xinvert : bool, optional
        If True, inverts the x-axis.
    yinvert : bool, optional
        If True, inverts the y-axis.
    xlog : bool, optional
        If True, the scale of the x-axis is logarithmic.
    ylog : bool, optional
        If True, the scale of the x-axis is logarithmic.
    title : str, optional
        Sets the title of the plot
    labels : array-like or boolean, optional
        Specifies label(s) for the contour(s). If given as an array of strings, sets the labels for each contour.
        Must be of equal length to number specified by 'percent'. If False, no legend is drawn.
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    lab_loc : int, optional
        Defines the position of the legend
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    output : boolean, optional
        If True, returns the edges and values of the underlying histogram in addition to the default output of.
    plot_kw : dict, optional
        Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are QuadContourSet properties.
    **kwargs: QuadContourSet properties, optional
        kwargs are used to specify matplotlib specific properties such as cmap, linewidths, hatches, etc.
        The list of available properties can be found here:
        https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.contour.html

    Returns
    -------
    bin_edges_x : array
        The bin edges for the x axis.
    bin_edges_y : array
        The bin edges for the y axis.
    n : array
        The values of the underlying histogram.
    l : array
        The levels for the contours.
    """

    from warnings import warn
    from matplotlib import lines, patches, rcParams
    from matplotlib.cm import get_cmap
    from matplotlib.pyplot import gca, contour, contourf, legend  # , Normalize
    from numpy import array, linspace, round, ndarray
    from scipy.ndimage.filters import gaussian_filter

    from .base_func import axes_handler, basehist2D, percent_finder, plot_finalizer
    from .defaults import Params

    # Check deprecated parameters
    if 'plabel' in kwargs.keys():
        warn("'plabel' will be deprecated, use instead 'labels'", DeprecationWarning, stacklevel=2)
        labels = kwargs.pop('plabel')

    # Initialise defaults
    if filled is None:
        filled = Params.cont_filled
    if percent is None:
        percent = Params.contp_percent
    if 'cmap' not in kwargs.keys() and 'cmap' not in plot_kw.keys() and 'colors' not in kwargs.keys() and 'colors' not in plot_kw.keys():
        plot_kw['cmap'] = Params.cont_cmap
    if output is None:
        output = Params.contp_output

    # Set the current axis
    if ax is not None:
        old_axes = axes_handler(ax)
    else:
        ax = gca()
        old_axes = ax

    if not isinstance(percent, ndarray):
        percent = array([percent]).flatten()

    if not isinstance(bin_type, (list, tuple, ndarray)):
        bin_type = [bin_type] * 2

    if not isinstance(bins, (list, tuple)):
        if bins is None:
            bins = max([10, int(len(x)**0.4)])  # Defaults to min of 10 bins
        bins = [bins] * 2

    percent = percent[::-1]  # reverse the order of percent

    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    # plot_par = {**plot_kw, **kwargs} # For Python > 3.5
    plot_par = plot_kw.copy()
    plot_par.update(kwargs)

    if filled:
        plot_par['extend'] = 'max'

    # if drawing <4 lines with no color specified, use 'solid', 'dashed' and then 'dotted'
    if not filled and len(percent) < 4 and 'colors' not in plot_par.keys():
        plot_par['colors'] = [next(ax._get_lines.prop_cycler)['color']] * len(percent)

    if 'colors' in plot_par.keys():
        if isinstance(plot_par['colors'], str):
            plot_par['colors'] = [plot_par['colors'] for i in range(len(percent))]

        if 'cmap' in plot_par.keys():
            plot_par.pop('cmap')
    elif max_spacing:
        if isinstance(plot_par['cmap'], str):
            plot_par['cmap'] = get_cmap(plot_par['cmap'])

        plot_par['colors'] = [plot_par['cmap'](i) for i in linspace(0, 1, len(percent))]
        plot_par.pop('cmap')

    # if drawing <4 lines with no color specified, use 'solid', 'dashed' and then 'dotted'
    if not filled and len(percent) < 4 and 'linestyles' not in plot_par.keys():
        plot_par['linestyles'] = [['solid', 'dashed', 'dotted'][i] for i in range(len(percent))][::-1]

    # Validate labels array
    if isinstance(labels, (list, tuple, ndarray)):
        if (len(labels) != len(percent)):
            raise ValueError(f"Length of labels ({len(labels)}) does not match length of percent ({len(percent)}).")
    else:
        if labels is None:
            if rcParams['text.usetex']:
                labels = [f'{round(p,1)}\\%' for p in percent]
            else:
                labels = [f'{round(p,1)}%' for p in percent]

    X, Y, Z = basehist2D(x, y, c=None, bin_type=bin_type, bin_num=bins, norm=None, dens=None, cstat=None, xlog=xlog, ylog=ylog)
    X = (X[:-1] + X[1:]) / 2
    Y = (Y[:-1] + Y[1:]) / 2

    level = array([percent_finder(Z, p / 100) for p in percent])
    if filled:  # Removed `func_dict` for readability
        contourset = contourf(X, Y, gaussian_filter(Z.T, sigma=smooth), levels=level, **plot_par)
    else:
        contourset = contour(X, Y, gaussian_filter(Z.T, sigma=smooth), levels=level, **plot_par)

    if labels:
        plot_par['colors'] = contourset.colors
        if isinstance(plot_par['colors'], str):
            plot_par['colors'] = [plot_par['colors']] * len(percent)

        if contourset.linestyles is None:
            plot_par['linestyles'] = ['solid'] * len(percent)
        elif isinstance(contourset.linestyles, str):
            plot_par['linestyles'] = [contourset.linestyles] * len(percent)
        else:
            plot_par['linestyles'] = contourset.linestyles

        if contourset.alpha is None:
            plot_par['alpha'] = [1.0 for i in range(len(percent))]
        elif isinstance(contourset.alpha, (float, int)):
            plot_par['alpha'] = [contourset.alpha for i in range(len(percent))]
        else:
            plot_par['alpha'] = contourset.alpha

        if filled:
            legend([patches.Patch(color=plot_par['colors'][i], alpha=plot_par['alpha'][i]) for i in range(len(percent))],
                   labels, numpoints=1, loc=lab_loc)
        else:
            legend([lines.Line2D([0, 1], [0, 1], color=plot_par['colors'][i], linestyle=plot_par['linestyles'][i], alpha=plot_par['alpha'][i]) for i in range(len(percent))],
                   labels, numpoints=1, loc=lab_loc)

    plot_finalizer(xlog, ylog, xlim, ylim, title, xlabel, ylabel, xinvert, yinvert, grid)
    if old_axes is not ax:
        old_axes = axes_handler(old_axes)

    if output:
        return(contourset, X, Y, Z.T)
    else:
        return(contourset)


####################################
##########  Error bands  ###########
####################################
def errorband(x, y, yerr, line=False, xlim=None, ylim=None,
              xinvert=False, yinvert=False, xlog=False, ylog=None, title=None, xlabel=None, ylabel=None,
              label=None, lab_loc=0, ax=None, grid=None, line_kw={}, band_kw={}, **kwargs):
    """Error line and band plotting function.

    Parameters
    ----------
    x : array-like or list
        If list it is assumed that each elemement is array-like.
    y : array-like or list
        If list it is assumed that each elemement is array-like.
    yerr : array-like or list, optional
        Defines the length of the errobars in the y-axis. If list it is assumed that each elemement is array-like.
    line : boolean, optional
        If True, draw a line that follows the statistic defined in line_stat.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    xinvert : bool or list, optional
        If True, inverts the x-axis.
    yinvert : bool or list, optional
        If True, inverts the y-axis.
    xlog : bool or list, optional
        If True, the scale of the x-axis is logarithmic.
    ylog : bool or list, optional
        If True, the scale of the x-axis is logarithmic.
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    label : str, optional
        Sets the label for the plot.
    lab_loc : int, optional
        Defines the position of the legend
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    plot_kw : dict, optional
        Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are Line2D properties.
    **kwargs: Line2D properties, optional
        kwargs are used to specify matplotlib specific properties such as linecolor, linewidth, antialiasing, etc.
        A list of available `Line2D` properties can be found here:
        https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D

    Returns
    -------
    None
    """

    from splotch.base_func import axes_handler,bin_axis,plot_finalizer

    import numpy as np
    from numbers import Number
    import scipy.stats as stats
    from numpy import array,percentile
    from functools import partial
    import matplotlib.colors as clr
    import matplotlib.pyplot as plt
    from matplotlib.pyplot import gca
    
    # Set the current axis
    if ax is not None:
        old_axes=axes_handler(ax)
    else:
        ax=gca()
        old_axes=ax
    
    if ylog is None:
        from splotch.defaults import Params
        ylog = Params.hist1D_yaxis_log
    if 'linewidth' not in band_kw.keys():
        band_kw['linewidth'] = 0
    if 'alpha' not in band_kw.keys():
        band_kw['alpha'] = 0.4

    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    # band_par={**plot_kw, **kwargs} # For Python > 3.5
    band_kw.update(kwargs)

    if len(array(yerr).shape) == 2:
        plt.fill_between(x, y - yerr[0], y + yerr[1], label=label, **band_kw)
    else:
        plt.fill_between(x, y - yerr, y + yerr, label=label, **band_kw)

    if line:
        plt.plot(x, y, **line_kw)
    if label is not None:
        plt.legend(loc=lab_loc)

    plot_finalizer(xlog, ylog, xlim, ylim, title, xlabel, ylabel, xinvert, yinvert, grid)
    if old_axes is not ax:
        old_axes = axes_handler(old_axes)

####################################
# Error bars
####################################
def errorbar(x, y, xerr=None, yerr=None, xlim=None, ylim=None, xinvert=False, yinvert=False, xlog=False, ylog=False,
             title=None, xlabel=None, ylabel=None, label=None, lab_loc=0, ax=None, grid=None, plot_kw={}, **kwargs):
    """Errorbar plotting function.

    This is a wrapper for pyplot.errorbar().

    Parameters
    ----------
    x : array-like or list
        If list it is assumed that each elemement is array-like.
    y : array-like or list
        If list it is assumed that each elemement is array-like.
    xerr : array-like or list, optional
        Defines the length of the errobars in the x-axis. If list it is assumed that each elemement is array-like.
    yerr : array-like or list, optional
        Defines the length of the errobars in the y-axis. If list it is assumed that each elemement is array-like.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    xinvert : bool or list, optional
        If True, inverts the x-axis.
    yinvert : bool or list, optional
        If True, inverts the y-axis.
    xlog : bool or list, optional
        If True, the scale of the x-axis is logarithmic.
    ylog : bool or list, optional
        If True, the scale of the x-axis is logarithmic.
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    label : str, optional
        Sets the label for the plot.
    lab_loc : int, optional
        Defines the position of the legend
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    plot_kw : dict, optional
        Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are Line2D properties.
    **kwargs: Line2D properties, optional
        kwargs are used to specify matplotlib specific properties such as linecolor, linewidth, antialiasing, etc.
        A list of available `Line2D` properties can be found here:
        https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D

    Returns
    -------
    None
    """

    from .base_func import axes_handler, dict_splicer, plot_finalizer

    from matplotlib.pyplot import errorbar, legend, gca
    
    # Set the current axis
    if ax is not None:
        old_axes=axes_handler(ax)
    else:
        ax=gca()
        old_axes=ax
    
    # Convert data to lists if needed
    if not isinstance(x, list): x = [x]
    if not isinstance(y, list): y = [y]
    if not isinstance(xerr, list): xerr = [xerr]
    if not isinstance(yerr, list): yerr = [yerr]

    L = len(x)
    if not isinstance(label, list):
        label = [label for i in range(L)]

    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    # plot_par={**plot_kw, **kwargs} # For Python > 3.5
    plot_par = plot_kw.copy()
    plot_par.update(kwargs)

    # Create 'L' number of plot kwarg dictionaries to parse into each plot call
    plot_par = dict_splicer(plot_par, L, [1] * L)

    for i in range(L):
        errorbar(x[i], y[i], xerr=xerr[i], yerr=yerr[i], label=label[i], **plot_par[i])
    if any(label):
        legend(loc=lab_loc)
    plot_finalizer(xlog, ylog, xlim, ylim, title, xlabel, ylabel, xinvert, yinvert, grid)
    if old_axes is not ax:
        old_axes = axes_handler(old_axes)

####################################
###########  Error boxes  ##########
####################################
def errorbox(x, y, xerr=None, yerr=None, xlim=None, ylim=None, xinvert=False, yinvert=False, xlog=False, ylog=False, box_type='ellipse',
             title=None, xlabel=None, ylabel=None, label=None, grid=None, lab_loc=0, ax=None, plot_kw={}, **kwargs):
    """Errorbox plotting function.

    This is a wrapper around matplotlib PatchCollections with a matplotlib errorbar functionality.

    Parameters
    ----------
    x : array-like or list
        If list it is assumed that each elemement is array-like.
    y : array-like or list
        If list it is assumed that each elemement is array-like.
    xerr : array-like or list, optional
        Defines the length of the errobars in the x-axis. If list it is assumed that each elemement is array-like.
    yerr : array-like or list, optional
        Defines the length of the errobars in the y-axis. If list it is assumed that each elemement is array-like.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    xinvert : bool or list, optional
        If True, inverts the x-axis.
    yinvert : bool or list, optional
        If True, inverts the y-axis.
    xlog : bool or list, optional
        If True, the scale of the x-axis is logarithmic.
    ylog : bool or list, optional
        If True, the scale of the x-axis is logarithmic.
    box_type : str
        The type of box to plot, patch types include: ellipse | rectangle (Default: ellipse).
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    label : str, optional
        Sets the label for the plot.
    lab_loc : int, optional
        Defines the position of the legend
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    plot_kw : dict, optional
        Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are Patches properties.
    **kwargs: Patch properties, optional
        kwargs are used to specify matplotlib specific properties such as facecolor, linestyle, alpha, etc.
        A list of available `Patch` properties can be found here:
        https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.patches.Rectangle.html

    Returns
    -------
    None
    """

    from .base_func import axes_handler, dict_splicer, plot_finalizer

    from matplotlib.pyplot import errorbar, legend, gca
    from numpy import shape, full, array
    from matplotlib.collections import PatchCollection
    from matplotlib.patches import Ellipse, Rectangle, Patch
    
    # Set the current axis
    if ax is not None:
        old_axes=axes_handler(ax)
    else:
        ax=gca()
        old_axes=ax
    
    if not isinstance(x, list): x = [x]
    if not isinstance(y, list): y = [y]
    if not isinstance(xerr, list): xerr = [xerr]
    if not isinstance(yerr, list): yerr = [yerr]

    boxdict = {'rec': Rectangle, 'ell': Ellipse}
    if (box_type.lower()[:3] not in ['rec', 'ell']):
        raise ValueError(f"box_type '{box_type}' not recognised.")

    L = len(x)
    if not isinstance(label, list):
        label = [label for i in range(L)]

    # Validate format of xerr and yerr
    for i in range(L):
        # x-axis errors
        if (shape(xerr[i]) == ()):  # Single error for all points
            xerr[i] = full((2, len(x[i])), xerr[i])
        else:
            if (len(shape(xerr[i])) == 1):
                if (shape(xerr[i])[0] == len(x[i])):  # single error for each point
                    xerr[i] = array([xerr[i], xerr[i]])
                elif (shape(xerr[i])[0] == 2):  # separate upper and lower errors for all points
                    xerr[i] = full((len(x[i]), 2), xerr[i]).T
                else:
                    raise ValueError(f"Invalid shape ({shape(xerr[i])}) for 'xerr' array.")
            elif (len(shape(xerr[i])) == 2):  # separate upper and lower errors for each point
                xerr[i] = array(xerr[i])
                if (shape(xerr[i])[0] != 2 or shape(xerr[i])[1] != len(x[i])):
                    raise ValueError(f"Invalid shape ({shape(xerr[i])}) for 'xerr' array.")

        # y-axis errors
        if (shape(yerr[i]) == ()):  # single error for all points
            yerr[i] = full((2, len(y[i])), yerr[i])
        else:
            if (len(shape(yerr[i])) == 1):
                if (shape(yerr[i])[0] == len(y[i])):  # single error for each point
                    yerr[i] = array([yerr[i], yerr[i]])
                elif (shape(yerr[i])[0] == 2):  # separate upper and lower errors for all points
                    yerr[i] = full((len(y[i]), 2), yerr[i]).T
                else:
                    raise ValueError(f"Invalid shape ({shape(yerr[i])}) for 'yerr' array.")
            elif (len(shape(yerr[i])) == 2):  # separate upper and lower errors for each point
                yerr[i] = array(yerr[i])
                if (shape(yerr[i])[0] != 2 or shape(yerr[i])[1] != len(y[i])):
                    raise ValueError(f"Invalid shape ({shape(yerr[i])}) for 'yerr' array.")

    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    # plot_par={**plot_kw, **kwargs} # For Python > 3.5
    plot_par = plot_kw.copy()
    plot_par.update(kwargs)

    # Create 'L' number of plot kwarg dictionaries to parse into each plot call
    plot_par = dict_splicer(plot_par, L, [1] * L)

    # Loop over data points; create box/ellipse from errors at each point
    boxdict = {'rec': Rectangle, 'ell': Ellipse}
    boxhandles = []
    for i in range(L):
        errorboxes = []
        for j, (xx, yy, xe, ye) in enumerate(zip(x[i], y[i], xerr[i].T, yerr[i].T)):
            errorboxes.append(boxdict[box_type.lower()[:3]]((xx - xe[0], yy - ye[0]), xe.sum(), ye.sum()))

        # Create and add patch collection with specified colour/alpha
        pc = PatchCollection(errorboxes, **plot_par[i])
        boxhandles.append(Patch(**plot_par[i]))
        ax.add_collection(pc)

    if any(label):
        legend(handles=boxhandles, labels=label, loc=lab_loc)
    plot_finalizer(xlog, ylog, xlim, ylim, title, xlabel, ylabel, xinvert, yinvert, grid)
    if old_axes is not ax:
        old_axes = axes_handler(old_axes)

####################################
# Hexagonal 2D histogram
####################################
def hexbin(x, y, bins=None, binlims=None, dens=True, scale=None,
           c=None, cstat=None, xlim=None, ylim=None, clim=[None, None], nmin=0,
           xinvert=False, yinvert=False, cbar_invert=False, xlog=False, ylog=False, clog=None, title=None, xlabel=None,
           ylabel=None, clabel=None, lab_loc=0, ax=None, grid=None, output=None, plot_kw={}, **kwargs):
    """Hexagonal 2D bins function.

    Parameters
    ----------
    x : array-like
        Position of data points in the x axis.
    y : array-like
        Position of data points in the y axis.
    bins : int or list, optional
        Gives the number of bins
    binlims : array-like, optional
        Defines the limits for the bin range, given as a one-dimensional list containing four elements (left, right, bottom, top).
        Note, if xlog=True or ylog=True, then the values of binlim must also be logged. The default assigns the limits based on
        gridsize, x, y, xscale and yscale.
    dens : bool or list, optional
        If false the histogram returns raw counts.
    c : array-like, optional
        If a valid argument is given in cstat, defines the value used for the binned statistics.
    cstat : str or function, optional
        Must be one of the valid str arguments for the statistics variable in scipy.stats.binned_statistic_2d
        ('mean’, 'median’, 'count’, 'sum’, 'min’ or 'max’) or a function that takes a 1D array and
        outputs an integer or float.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    clim : list, optional
        Defines the limits of the colour map ranges, it must contain two elements (lower and higer limits).
    nmin : int, optional (default: 0)
        The minimum number of points required in a bin in order to be plotted.
    xinvert : bool, optional
        If True, inverts the x-axis.
    yinvert : bool, optional
        If True, inverts the y-axis.
    cbar_invert : bool, optional
        If True, inverts the direction of the colour bar (not the colour map).
    xlog : bool, optional
        If True, the scale of the x-axis is logarithmic.
    ylog : bool, optional
        If True, the scale of the x-axis is logarithmic.
    clog : bool, optional
        If True, the colour map is changed from linear to logarithmic.
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    clabel : str, optional
        Setting `clabel` triggers the generation of a colourbar with axis label given by its value.
    lab_loc : int, optional
        Defines the position of the legend
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    output : boolean, optional
        If True, returns the edges and values of the histogram.
    plot_kw : dict, optional
        Explicit dictionary of kwargs to be parsed to matplotlib hexbin function.
        Parameters will be overwritten if also given implicitly as a **kwarg.
    **kwargs : pcolormesh properties, optional
        kwargs are used to specify matplotlib specific properties such as cmap, norm, edgecolors etc.
        https://matplotlib.org/api/_as_gen/matplotlib.pyplot.hexbin.html

    Returns
    -------
    n : array
        The values of the histogram. Only provided if output is True.
    x_edges : array
        The bin edges for the x axis. Only provided if output is True.
    y_edges : array
        The bin edges for the y axis. Only provided if output is True.
    """
    
    from numpy import diff, log10, nan, nanmin, nanmedian, nanmax, nanstd, unique, size, zeros, shape
    from matplotlib.colors import LogNorm
    from matplotlib.pyplot import gca, hexbin, colorbar
    from .base_func import axes_handler, plot_finalizer
    
    # Set the current axis
    if ax is not None:
        old_axes=axes_handler(ax)
    else:
        ax=gca()
        old_axes=ax
    
    if type(bins) not in [list, tuple]:
        if bins is None:
            bins = max([10, int(len(x)**0.4)])  # Defaults to min of 10 bins
        bins = [bins, int(bins / (3**0.5))]

    if None in (clog, output):
        from .defaults import Params
        if clog is None:
            clog = Params.hist2D_caxis_log
        if output is None:
            output = Params.hist2D_output

    if size([x, y]) == 0:  # Zero-sized arrays given
        if (clog is True): raise ValueError("Cannot set 'clog'=True if zero-size array given.")
        if (cstat is not None): raise ValueError(f"Cannot compute statistic (cstat='{cstat}') on zero-size array, set cstat=None if no data given.")

    # Create a temporary duplicate of x and y data
    temp_x = x * 1.0  # this should probably use: copy.deepcopy(x)
    temp_y = y * 1.0

    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    # plot_par = {**plot_kw, **kwargs}  # For Python > 3.5
    plot_par = plot_kw.copy()
    plot_par.update(kwargs)
    if binlims:
        plot_par['extent'] = binlims

    if c is not None:
        plot_par['C'] = c
    if cstat:
        cstat_func = {'min': nanmin, 'mean': nanmax, 'median': nanmedian, 'max': nanmax, 'std': nanstd}
        if cstat in cstat_func.keys():
            plot_par['reduce_C_function'] = cstat_func[cstat]
        else:
            plot_par['reduce_C_function'] = cstat

    if clim:
        plot_par['vmin'] = clim[0]
        plot_par['vmax'] = clim[1]

    if nmin:
        plot_par['mincnt'] = nmin

    if xlog:
        plot_par['xscale'] = 'log'
        temp_x = log10(temp_x)

    if ylog:
        plot_par['yscale'] = 'log'
        temp_y = log10(temp_y)

    if clog:
        plot_par['bins'] = 'log'
        if 'mincnt' not in plot_par.keys():
            plot_par['mincnt'] = 1

    if dens and c is None:
        # This is nasty, but seems to be the quickest way to do this without fully rewriting hexbin here
        hist_return = hexbin(temp_x, temp_y, gridsize=bins)
        hist_return.remove()
        offsets = hist_return.get_offsets()
        offsets_x = unique(offsets[:, 0])
        offsets_y = unique(offsets[:, 1])
        hex_area = diff(offsets_x)[0] * 2 * diff(offsets_y)[0]

        def density_scaling(bin_data):
            bin_dens = 1.0 * len(bin_data) / (hex_area)
            if scale:
                bin_dens /= 1.0 * scale
            else:
                bin_dens /= len(x)
            return(bin_dens)

        plot_par['C'] = y
        plot_par['reduce_C_function'] = density_scaling

    hist_return = hexbin(x, y, gridsize=bins, **plot_par)

    if clabel is not None:
        cbar = colorbar()
        cbar.set_label(clabel)
        if cbar_invert:
            cbar.ax.invert_yaxis()

    plot_finalizer(xlog, ylog, xlim, ylim, title, xlabel, ylabel, xinvert, yinvert, grid)

    if old_axes is not ax:
        old_axes = axes_handler(old_axes)
    if output:
        return(hist_return.get_array(), hist_return.get_offsets())

########################################
## 2D histogram and binned statistics ##
########################################
def hist2D(x, y, bin_type=None, bins=None, dens=True, scale=None, c=None, cstat=None, xlim=None, ylim=None, clim=[None, None], nmin=0,
           xinvert=False, yinvert=False, cbar_invert=False, xlog=False, ylog=False, clog=None, title=None, xlabel=None,
           ylabel=None, clabel=None, lab_loc=0, ax=None, grid=None, output=None, plot_kw={}, **kwargs):
    """2D histogram function.

    Parameters
    ----------
    x : array-like
        Position of data points in the x axis.
    y : array-like
        Position of data points in the y axis.
    bin_type : {'number','width','edges','equal'}, optional
        Defines how is understood the value given in bins: 'number' for givinf the desired number of
        bins, 'width' for the width of the bins, 'edges' for the edges of bins, and 'equal' for
        making bins with equal number of elements (or as close as possible). If not given it is
        inferred from the data type of bins: 'number' if int, 'width' if float and 'edges' if ndarray.
    bins : int, float, array-like or list, optional
        Gives the values for the bins, according to bin_type.
    dens : bool or list, optional
        If false the histogram returns raw counts.
    scale : float or list, optional
        Scaling of the data counts.
    c : array-like, optional
        If a valid argument is given in cstat, defines the value used for the binned statistics.
    cstat : str or function, optional
        Must be one of the valid str arguments for the statistics variable in scipy.stats.binned_statistic_2d
        ('mean’, 'median’, 'count’, 'sum’, 'min’ or 'max’) or a function that takes a 1D array and
        outputs an integer or float.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    clim : list, optional
        Defines the limits of the colour map ranges, it must contain two elements (lower and higer limits).
    nmin : int, optional (default: 0)
        The minimum number of points required in a bin in order to be plotted.
    xinvert : bool, optional
        If True, inverts the x-axis.
    yinvert : bool, optional
        If True, inverts the y-axis.
    cbar_invert : bool, optional
        If True, inverts the direction of the colour bar (not the colour map).
    xlog : bool, optional
        If True, the scale of the x-axis is logarithmic.
    ylog : bool, optional
        If True, the scale of the x-axis is logarithmic.
    clog : bool, optional
        If True, the colour map is changed from linear to logarithmic.
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    clabel : str, optional
        Setting `clabel` triggers the generation of a colourbar with axis label given by its value.
    lab_loc : int, optional
        Defines the position of the legend
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    output : boolean, optional
        If True, returns the edges and values of the histogram.
    plot_kw : dict, optional
        Explicit dictionary of kwargs to be parsed to matplotlib pcolormesh function.
        Parameters will be overwritten if also given implicitly as a **kwarg.
    **kwargs : pcolormesh properties, optional
        kwargs are used to specify matplotlib specific properties such as cmap, norm, edgecolors etc.
        https://matplotlib.org/api/_as_gen/matplotlib.pyplot.pcolormesh.html

    Returns
    -------
    n : array
        The values of the histogram. Only provided if output is True.
    x_edges : array
        The bin edges for the x axis. Only provided if output is True.
    y_edges : array
        The bin edges for the y axis. Only provided if output is True.
    """

    from numpy import nan, size, zeros, shape
    from matplotlib.colors import LogNorm
    from matplotlib.pyplot import gca, pcolormesh, colorbar
    from .base_func import axes_handler, basehist2D, plot_finalizer
    
    # Set the current axis
    if ax is not None:
        old_axes=axes_handler(ax)
    else:
        ax=gca()
        old_axes=ax
    
    if not isinstance(bin_type, list):
        bin_type = [bin_type] * 2
    if type(bins) not in [list, tuple]:
        if bins is None:
            bins = max([10, int(len(x)**0.4)])  # Defaults to min of 10 bins
        bins = [bins] * 2

    if None in (clog, output):
        from .defaults import Params
        if clog is None:
            clog = Params.hist2D_caxis_log
        if output is None:
            output = Params.hist2D_output

    if size([x, y]) == 0:  # Zero-sized arrays given
        if (clog is True): raise ValueError("Cannot set 'clog'=True if zero-size array given.")
        if (cstat is not None): raise ValueError(f"Cannot compute statistic (cstat='{cstat}') on zero-size array, set cstat=None if no data given.")

    X, Y, Z = basehist2D(x, y, c, bin_type, bins, scale, dens, cstat, xlog, ylog)

    # Also get counts for number threshold cut
    if (size([x, y]) == 0):
        counts = zeros(shape=shape(Z))
    else:
        _, _, counts = basehist2D(x, y, c, bin_type, bins, None, False, None, xlog, ylog)

    # Cut bins which do not meet the number count threshold
    Z[counts < nmin] = nan

    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    # plot_par = {**plot_kw, **kwargs}  # For Python > 3.5
    plot_par = plot_kw.copy()
    plot_par.update(kwargs)

    if clog:
        pcolormesh(X, Y, Z.T, norm=LogNorm(vmin=clim[0], vmax=clim[1], clip=False), **plot_par)
    else:
        pcolormesh(X, Y, Z.T, vmin=clim[0], vmax=clim[1], **plot_par)

    if clabel is not None:
        cbar = colorbar()
        cbar.set_label(clabel)
        if cbar_invert:
            cbar.ax.invert_yaxis()
    plot_finalizer(xlog, ylog, xlim, ylim, title, xlabel, ylabel, xinvert, yinvert, grid)
    if old_axes is not ax:
        old_axes = axes_handler(old_axes)
    if output:
        return(Z.T, X, Y)


####################################
######  Image from 2D array  #######
####################################
def img(im, x=None, y=None, xlim=None, ylim=None, clim=[None, None], cmin=0, xinvert=False, yinvert=False, cbar_invert=False, clog=None, 
        title=None, xlabel=None, ylabel=None, clabel=None, lab_loc=0, ax=None, grid=None, plot_kw={}, **kwargs):

    """2D pixel-based image plotting function.

    Parameters
    ----------
    im : array-like
        Value for each pixel in an x-y 2D array, where the first dimension is the x-position and the
        second is the y-position.
    x : array-like, optional
        Position of data points in the x axis.
    y : array-like, optional
        Position of data points in the y axis.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    clim : list, optional
        Defines the limits of the colour map ranges, it must contain two elements (lower and higer limits).
    clog : bool, optional
        If True, the colour map is changed from linear to logarithmic.
    xinvert : bool, optional
        If True, inverts the x-axis.
    yinvert : bool, optional
        If True, inverts the y-axis.
    cbar_invert : bool, optional
        If True, inverts the direction of the colour bar (not the colour map).
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    clabel : str, optional
        Setting `clabel` triggers the generation of a colourbar with axis label given by its value.
    lab_loc : int, optional
        Defines the position of the legend
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    plot_kw : dict, optional
        Explicit dictionary of kwargs to be parsed to matplotlib pcolormesh function.
        Parameters will be overwritten if also given implicitly as a **kwarg.
    **kwargs : pcolormesh properties, optional
        kwargs are used to specify matplotlib specific properties such as `cmap`, `marker`, `norm`, etc.
        A list of available `pcolormesh` properties can be found here:
        https://matplotlib.org/api/_as_gen/matplotlib.pyplot.pcolormesh.html

    Returns
    -------
    None
    """

    from numpy import arange, meshgrid
    from matplotlib.colors import LogNorm
    from matplotlib.pyplot import gca, pcolormesh, colorbar
    from .base_func import axes_handler, plot_finalizer
    
    # Set the current axis
    if ax is not None:
        old_axes=axes_handler(ax)
    else:
        ax=gca()
        old_axes=ax
    
    if x is None:
        x = arange(len(im[:, 0]) + 1)
    if y is None:
        y = arange(len(im[0, :]) + 1)
    if clog is None:
        from .defaults import Params
        clog = Params.img_caxis_log

    X, Y = meshgrid(x, y)

    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    # plot_par = {**plot_kw, **kwargs}  # For Python > 3.5
    plot_par = plot_kw.copy()
    plot_par.update(kwargs)

    if clog:
        pcolormesh(X, Y, im.T, norm=LogNorm(vmin=clim[0], vmax=clim[1], clip=True), **plot_par)
    else:
        pcolormesh(X, Y, im.T, vmin=clim[0], vmax=clim[1], **plot_par)

    if clabel is not None:
        cbar = colorbar()
        cbar.set_label(clabel)
        if cbar_invert:
            cbar.ax.invert_yaxis()

    plot_finalizer(False, False, xlim, ylim, title, xlabel, ylabel, xinvert, yinvert, grid)

    if old_axes is not ax:
        old_axes = axes_handler(old_axes)


####################################
##########  Scatter plots ##########
####################################
def scatter(x, y, c=None, xlim=None, ylim=None, clim=None, density=False, xinvert=False, yinvert=False, cbar_invert=False, xlog=False, ylog=False, title=None,
            xlabel=None, ylabel=None, clabel=None, label=None, lab_loc=0, ax=None, grid=None, plot_kw={}, **kwargs):
    """2D pixel-based image plotting function.

    Parameters
    ----------
    x : array-like or list
        Position of data points in the x-axis.
    y : array-like or list
        Position of data points in the y-axis.
    c : array-like or list or str, optional
        Value of data points in the z-axis (colour-axis).
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    clim : tuple-like, optional
        Defines the limits of the colour-axis, it must contain two elements (lower and higer limits).
        Functions equivalently to the `vmin, vmax` arguments used by `colors.Normalize`. If both are
        given, `clim` takes priority.
    density : bool, optional
        If True, color-codes points by their spatial density to nearby points using a Gaussian
        kernel density estimate. If 'c' also given, 'density' takes precedence. Default: False.
    xinvert : bool, optional
        If True, inverts the x-axis.
    yinvert : bool, optional
        If True, inverts the y-axis.
    cbar_invert : bool, optional
        If True, inverts the direction of the colour bar (not the colour map).
    xlog : bool, optional
        If True, the scale of the x-axis is logarithmic.
    ylog : bool, optional
        If True, the scale of the x-axis is logarithmic.
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    clabel : str, optional
        Setting `clabel` triggers the generation of a colourbar with axis label given by its value.
    label : str, optional
        Sets the label for the scatter plot.
    lab_loc : int, optional
        Defines the position of the legend
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    plot_kw : dict, optional
        Explicit dictionary of kwargs to be parsed to matplotlib scatter function.
        Parameters will be overwritten if also given implicitly as a **kwarg.
    **kwargs : Collection properties, optional
        kwargs are used to specify matplotlib specific properties such as cmap, marker, norm, etc.
        A list of available `Collection` properties can be found here:
        https://matplotlib.org/api/collections_api.html#matplotlib.collections.Collection

    Returns
    -------
    paths
        A list of PathCollection objects representing the plotted data.
    """

    from numpy import array, dtype, shape, vstack
    from matplotlib.pyplot import gca, scatter, legend
    from scipy.stats import gaussian_kde
    from warnings import warn

    from .base_func import axes_handler, dict_splicer, plot_finalizer
    from .axis_func import colorbar
    
    # Set the current axis
    if ax is not None:
        old_axes=axes_handler(ax)
    else:
        ax=gca()
        old_axes=ax
    
    if (not isinstance(x, list)) or (len(shape(x)) == 1 and array(x).dtype is not dtype('O')):
        x = [x]
    if (not isinstance(y, list)) or (len(shape(y)) == 1 and array(y).dtype is not dtype('O')):
        y = [y]

    L = len(x)

    if (not isinstance(c, list)) or (len(shape(c)) == 1 and array(c).dtype is not dtype('O')):
        c = [c]
        if isinstance(c[0], str) or c[0] is None:
            c = [c[0] for i in range(L)]
    if not isinstance(label, list):
        label = [label for i in range(L)]

    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    # plot_par = {**plot_kw, **kwargs}  # For Python > 3.5
    plot_par = plot_kw.copy()
    plot_par.update(kwargs)

    # Insert clim as vmin, vmax into **kwargs dictionary, if given.
    if (clim is not None):
        try:
            _ = (e for e in clim)
            if (len(clim) == 2):
                plot_par['vmin'] = clim[0]
                plot_par['vmax'] = clim[1]
            else:
                raise TypeError("`clim` must be of iterable type and have two values only.")
        except (TypeError):
            raise TypeError("`clim` must be of iterable type and have two values only.")

    if (density is True):
        if (all([kk is not None for kk in c])):
            warn("Cannot specify both `c` and `density`, ignoring `c`.")
        c = [None] * L

        for i in range(L):
            xy = vstack([x[i], y[i]])
            c[i] = gaussian_kde(xy)(xy)  # Calculate the Gaussian kernel density estimate

    # Create 'L' number of plot kwarg dictionaries to parse into each scatter call
    plot_par = dict_splicer(plot_par, L, [len(i) for i in x])

    paths = []
    for i in range(L):
        p = scatter(x[i], y[i], c=c[i], label=label[i], **plot_par[i])
        paths.append(p)
    if clabel is not None:
        cbar = colorbar()
        cbar.set_label(clabel)
        if cbar_invert:
            cbar.ax.invert_yaxis()
    if any(label):
        legend(loc=lab_loc)

    plot_finalizer(xlog, ylog, xlim, ylim, title, xlabel, ylabel, xinvert, yinvert, grid)
    if old_axes is not ax:
        old_axes = axes_handler(old_axes)

    return paths[0] if len(paths) == 1 else paths


####################################
##########  Sector plots ###########
####################################
def sector(r, theta, rlim=(0.0, 1.0), thetalim=(0.0, 360.0), clim=None, rotate=0.0, rlabel="", thetalabel="", clabel=None, label=None, rstep=None,
           thetastep=15.0, rticks='auto', thetaticks='auto', cbar_invert=False, fig=None, plot_kw={}, **kwargs):
    """ Sector Plot function

    Plots a sector plot (a.k.a "pizza plot") based on data with one radial axis and an angular axis

    Parameters
    ----------
    r : array-like or list
        Radial axis data.
    theta : array-like or list
        Angular axis data (degrees).
    rlim : tuple-like, optional
        The lower and upper limits for the radial axis (degrees).
    thetalim : tuple-like, optional
        The lower and upper limits for the angular axis (degrees).
    clim : tuple-like, optional
        Defines the limits of the colour-axis, it must contain two elements (lower and higer limits).
        Functions equivalently to the `vmin, vmax` arguments used by `colors.Normalize`. If both are
        given, `clim` takes priority.
    rotate : float, optional
        By how many degrees (clockwise) to rotate the entire plot (valid values in [-180, 180]).
    rlabel : str, optional
        Sets the label of the r-axis.
    thetalabel : str, optional
        Sets the label of the theta-axis.
    clabel : str, optional
        Setting `clabel` triggers the generation of a colourbar with axis label given by its value.
    label : str, optional
        Sets the label for the scatter plot.
    rstep : float, optional
        Sets the step size of r ticks.
    thetastep : float, optional, default: 15.0
        Sets the step size of theta ticks (degrees).
    rticks : 'auto', or ticker
        * Not implement *
    thetaticks : 'auto', or ticker
        * Not implement *
    cbar_invert : bool, optional
        If True, inverts the direction of the colour bar (not the colour map).
    fig : pyplot.Figure, optional
        Use the given figure to make the plot, defaults to the current figure.
    plot_kw : dict, optional
        Explicit dictionary of kwargs to be parsed to matplotlib scatter function.
        Parameters will be overwritten if also given implicitly in **kwargs.
    **kwargs : Collection properties, optional
        kwargs are used to specify matplotlib specific properties such as cmap, marker, norm, etc.
        A list of available `Collection` properties can be found here:
        https://matplotlib.org/3.1.0/api/collections_api.html#matplotlib.collections.Collection

    Returns
    -------
    ax : The pyplot.Axes object created for the sector plot.
    """

    from matplotlib.transforms import Affine2D
    from matplotlib.projections.polar import PolarAxes
    from matplotlib.pyplot import gcf, colorbar, legend

    from mpl_toolkits.axisartist import floating_axes
    from mpl_toolkits.axisartist.grid_finder import (FixedLocator, MaxNLocator, DictFormatter)
    import mpl_toolkits.axisartist.angle_helper as angle_helper

    from numpy import array, linspace, arange, shape, sqrt, floor, round, degrees, radians, pi

    if (fig is None):
        fig = gcf()

    # rotate a bit for better orientation
    trans_rotate = Affine2D().translate(0.0, 0)

    # scale degree to radians
    trans_scale = Affine2D().scale(pi / 180.0, 1.0)
    trans = trans_rotate + trans_scale + PolarAxes.PolarTransform()

    # Get theta ticks
    thetaticks = arange(*radians(array(thetalim) - rotate), step=radians(thetastep))
    theta_gridloc = FixedLocator(thetaticks[thetaticks / (2 * pi) < 1])
    theta_tickfmtr = DictFormatter(dict(zip(thetaticks, [f"{(round(degrees(tck)+rotate)):g}" for tck in thetaticks])))

    # tick_fmtr=DictFormatter(dict(angle_ticks))
    # tick_fmtr=angle_helper.Formatter()

    if (rstep is None):
        rstep = 0.5

    r_gridloc = FixedLocator(arange(rlim[0], rlim[1], step=rstep))

    grid = floating_axes.GridHelperCurveLinear(
        PolarAxes.PolarTransform(),
        extremes=(*radians(array(thetalim) - rotate), *rlim),
        grid_locator1=theta_gridloc,
        grid_locator2=r_gridloc,
        tick_formatter1=theta_tickfmtr,
        tick_formatter2=None,
    )

    ax = floating_axes.FloatingSubplot(fig, 111, grid_helper=grid)
    fig.add_subplot(ax)

    # tick references
    thetadir_ref = ['top', 'right', 'bottom', 'left']
    rdir_ref = ['bottom', 'left', 'top', 'right']

    # adjust axes directions
    ax.axis["left"].set_axis_direction('bottom')  # Radius axis (displayed)
    ax.axis["right"].set_axis_direction('top')  # Radius axis (hidden)
    ax.axis["top"].set_axis_direction('bottom')  # Theta axis (outer)
    ax.axis["bottom"].set_axis_direction('top')  # Theta axis (inner)

    # Top theta axis
    ax.axis["top"].toggle(ticklabels=True, label=True)
    ax.axis["top"].major_ticklabels.set_axis_direction(thetadir_ref[(int(rotate) // 90) % 4])
    ax.axis["top"].label.set_axis_direction(thetadir_ref[(int(rotate) // 90) % 4])

    # Bottom theta axis
    ax.axis["bottom"].set_visible(False if rlim[0] < (rlim[1] - rlim[0]) / 3 else True)
    ax.axis["bottom"].major_ticklabels.set_axis_direction(thetadir_ref[(int(rotate) // 90 + 2) % 4])

    # Visible radius axis
    ax.axis["left"].major_ticklabels.set_axis_direction(rdir_ref[(int(rotate) // 90) % 4])
    ax.axis["left"].label.set_axis_direction(rdir_ref[(int(rotate) // 90) % 4])

    # Labels
    ax.axis["left"].label.set_text(rlabel)
    ax.axis["top"].label.set_text(thetalabel)

    # create a parasite axes whose transData in RA, cz
    sector_ax = ax.get_aux_axes(trans)

    # This has a side effect that the patch is drawn twice, and possibly over some other
    # artists. So, we decrease the zorder a bit to prevent this.
    sector_ax.patch = ax.patch
    sector_ax.patch.zorder = 0.9

    L = shape(theta)[0] if len(shape(theta)) > 1 else 1
    plot_par = plot_kw.copy()
    plot_par.update(kwargs)

    # Insert clim as vmin, vmax into **kwargs dictionary, if given.
    if (clim is not None):
        try:
            _ = (e for e in clim)
            if (len(clim) == 2):
                plot_par['vmin'] = clim[0]
                plot_par['vmax'] = clim[1]
            else:
                raise TypeError("`clim` must be of iterable type and have two values only.")
        except (TypeError):
            raise TypeError("`clim` must be of iterable type and have two values only.")

    # Create 'L' number of plot kwarg dictionaries to parse into each plot call
    # plot_par=dict_splicer(plot_par,L,[1]*L)

    if (L == 1):
        sctr = sector_ax.scatter(theta - rotate, r, label=label, **plot_par)
    else:
        for ii in range(L):
            sctr = sector_ax.scatter(theta[ii] - rotate, r[ii], label=label[ii], **plot_par[ii])

    if clabel is not None:
        cbar = colorbar(sctr)
        cbar.set_label(clabel)
        if cbar_invert:
            cbar.ax.invert_yaxis()

    return sector_ax


####################################
########  Statistics bands  ########
####################################
def statband(x, y, bin_type=None, bins=None, stat_mid='mean', stat_low='std', stat_high='std', from_mid=None, line=False, xlim=None, ylim=None,
             xinvert=False, yinvert=False, xlog=False, ylog=None, nmin=0, title=None, xlabel=None, ylabel=None,
             label=None, lab_loc=0, ax=None, grid=None, line_kw={}, band_kw={}, **kwargs):
    """Statistics line and band plotting function.

    Parameters
    ----------
    x : array-like or list
        If list it is assumed that each elemement is array-like.
    y : array-like or list
        If list it is assumed that each elemement is array-like.
    bin_type : {'number','width','edges','equal'}, optional
        Defines how is understood the value given in bins: 'number' for the desired number of bins, 'width' for the width
        of the bins, 'edges' for the edges of bins, and 'equal' for making bins with equal number of elements (or as close
        as possible). If not given it is inferred from the data type of bins: 'number' if int, 'width' if float and 'edges'
        if ndarray.
    bins : int, float, array-like or list, optional
        Gives the values for the bins, according to bin_type.
    stat_mid : str, int, float or function, optional
        Defines how to calculate the midpoint of the statistics band. When passing a string it must be either one of the options
        for scipy.stats.binned_statistic(), i.e. 'mean', 'std', 'median', 'count', 'sum', 'min', 'max' or a user-defined function.
        If given as an integer or float, the number represents the value for the percentile to calculate in each bin.
        A function can be given which takes (only) a 1D array of values and returns a numerical statistic.
    stat_low / stat_high : str, int, float or function, optional
        Defines how to calculate the lower/upper limits for the statistic band. Can be given as one of the recognised strings above or as
        a string combining 'std' with a number, i.e. '[n]std', where [n] is the number of standard deviations away from the line of `stat_mid`.
        Can also be given as a number (integer or float) or function as described for stat_mid.
    from_mid : boolean, optional
        If True, the lower/upper bounds of the band are determined as the separation from the stat_mid line: i.e. stat_mid +/- stat_[low/high],
        otherwise, they are set to the values returned by stat_[low/high]. Defaults to True if stat_[low/high] are standard deviations.
    line : boolean, optional
        If True, draw a line that follows the statistic defined in line_stat.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    xinvert : bool or list, optional
        If True, inverts the x-axis.
    yinvert : bool or list, optional
        If True, inverts the y-axis.
    xlog : bool or list, optional
        If True, the scale of the x-axis is logarithmic.
    ylog : bool or list, optional
        If True, the scale of the x-axis is logarithmic.
    nmin : int, optional (default: 0)
        The minimum number of points required in a bin in order to be plotted.
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    label : str, optional
        Sets the label for the plot.
    lab_loc : int, optional
        Defines the position of the legend
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    plot_kw : dict, optional
        Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are Line2D properties.
    **kwargs: Line2D properties, optional
        kwargs are used to specify matplotlib specific properties such as linecolor, linewidth, antialiasing, etc.
        A list of available `Line2D` properties can be found here:
        https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D

    Returns
    -------
    None
    """

    from splotch.base_func import axes_handler, bin_axis, plot_finalizer

    import numpy as np
    from numbers import Number
    import scipy.stats as stats
    from numpy import percentile
    from functools import partial
    import matplotlib.colors as clr
    import matplotlib.pyplot as plt
    from matplotlib.pyplot import gca
    
    # Set the current axis
    if ax is not None:
        old_axes=axes_handler(ax)
    else:
        ax=gca()
        old_axes=ax
    
    if ylog is None:
        from splotch.defaults import Params
        ylog = Params.hist1D_yaxis_log
    if bins is None:
        bins = int((len(x))**0.4)
    if 'linewidth' not in band_kw.keys():
        band_kw['linewidth'] = 0
    if 'alpha' not in band_kw.keys():
        band_kw['alpha'] = 0.4

    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    # band_par={**plot_kw, **kwargs} # For Python > 3.5
    band_kw.update(kwargs)

    # Check stat_low/stat_high arguments
    band_stat = np.array([None, None])
    band_multi = np.ones(2)
    for i, stat in enumerate([stat_low, stat_high]):  # loop over low/high statistic
        if isinstance(stat, Number):  # stat given as a percentile number
            band_stat[i] = partial(percentile, q=stat)  # Set as the percentile function with kwargs fixed.
        elif callable(stat):  # stat given as a function
            band_stat[i] = stat
        elif isinstance(stat, str) and 'std' in stat:  # stat given as a string with 'std'
            band_stat[i] = 'std'
            multi = list(filter(None, stat.split('std')))  # The multiplier for std, if any.
            if len(multi) == 0:  # No multiplier given
                band_multi[i] = 1.0
            elif len(multi) == 1:  # Multiplier was given
                band_multi[i] = multi[0]
            else:
                raise ValueError(f"Statistic '{stat}' not valid. Should be given as '[n]std', where [n] is a number.")
        else:
            raise ValueError(f"Statistic of type '{type(stat)}' was not recognised. Must be either a Number, function or string in the format '[n]std'.")

    # Check stat_mid argument
    if isinstance(stat_mid, Number):
        stat_mid = partial(percentile, q=stat_mid)

    # Assign 'from_mid' if not explicitly set
    if from_mid is None:
        if (band_stat[0] in ['std', np.std, np.nanstd] and (band_stat[1] in ['std', np.std, np.nanstd])):  # Band stats a type of standard deviation
            from_mid = True
        else:
            from_mid = False

    temp_x, bins_hist, bins_plot = bin_axis(x, bin_type, bins, log=xlog)
    temp_y = stats.binned_statistic(temp_x, y, statistic=stat_mid, bins=bins_hist)[0]
    if band_stat[0] == band_stat[1]:
        band_low, band_high = [stats.binned_statistic(temp_x, y, statistic=band_stat[0], bins=bins_hist)[0]] * 2
    else:
        band_low = stats.binned_statistic(temp_x, y, statistic=band_stat[0], bins=bins_hist)[0]
        band_high = stats.binned_statistic(temp_x, y, statistic=band_stat[1], bins=bins_hist)[0]

    if from_mid is True:  # Band intervals should be taken as the difference from the mid line
        band_low = temp_y - band_multi[0] * band_low
        band_high = temp_y + band_multi[1] * band_high

    if ylog:
        temp_y = np.where(temp_y == 0, np.nan, temp_y)
    
    counts = np.histogram(temp_x, bins=bins_hist)[0]
    x = stats.binned_statistic(temp_x, x, statistic=stat_mid, bins=bins_hist)[0]
    y = temp_y
    
    x = x[counts>nmin]
    y = y[counts>nmin]
    band_low = band_low[counts>nmin]
    band_high = band_high[counts>nmin]
    
    plt.fill_between(x, band_low, band_high, label=label, **band_kw)

    if line:
        plt.plot(x, y, **line_kw)
    if label is not None:
        plt.legend(loc=lab_loc)

    plot_finalizer(xlog, ylog, xlim, ylim, title, xlabel, ylabel, xinvert, yinvert, grid)
    if old_axes is not ax:
        old_axes = axes_handler(old_axes)


####################################
########  Statistics bars  #########
####################################
def statbar(x, y, bin_type=None, bins=None, stat_cen='mean', bar_x=True, stat_y='std', line=False, xlim=None, ylim=None,
            xinvert=False, yinvert=False, xlog=False, ylog=None, title=None, xlabel=None, ylabel=None,#nmin=0,
            label=None, lab_loc=0, ax=None, grid=None, plot_kw={}, **kwargs):
    """Statistics line and bar plotting function.

    Parameters
    ----------
    x : array-like or list
        If list it is assumed that each elemement is array-like.
    y : array-like or list
        If list it is assumed that each elemement is array-like.
    bin_type : {'number','width','edges','equal'}, optional
        Defines how is understood the value given in bins: 'number' for the desired number of bins, 'width' for the width
        of the bins, 'edges' for the edges of bins, and 'equal' for making bins with equal number of elements (or as close
        as possible). If not given it is inferred from the data type of bins: 'number' if int, 'width' if float and 'edges'
        if ndarray.
    bins : int, float, array-like or list, optional
        Gives the values for the bins, according to bin_type.
    stat_cen : str, int, float, function, or 2-element array-like of any of the previous type, optional
        Defines how to calculate the position of centre of each errorbar. When passing an integer or float is
        interpreted as being the percentile for the limit. When passing a function it must have the input and
        ouput characteristics required by scipy.stats.binned_statistic().
    bar_x : bool, optional
        If False turns off the display of the bin widths with bars.
    stat_y : str, int, float, function, or 2-element array-like of any of the previous type, optional
        Defines how to calculate the y error bars. When passing a string it must be either one of the options
        for scipy.stats.binned_statistic(), or a string that combines 'std' with a number (e.g., '2.2std'), where to number is
        interpreted as the number of standard deviations that the limit must cover. When passing an integer or float is
        interpreted as being the percentile for the limit. When passing a function it must have the input and ouput
        characteristics required by scipy.stats.binned_statistic().
    line : boolean, optional
        If True, draw a line that follows the statistic defined in line_stat.
    xlim : tuple-like, optional
        Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
    ylim : tuple-like, optional
        Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
    xinvert : bool or list, optional
        If True, inverts the x-axis.
    yinvert : bool or list, optional
        If True, inverts the y-axis.
    xlog : bool or list, optional
        If True, the scale of the x-axis is logarithmic.
    ylog : bool or list, optional
        If True, the scale of the x-axis is logarithmic.
    title : str, optional
        Sets the title of the plot
    xlabel : str, optional
        Sets the label of the x-axis.
    ylabel : str, optional
        Sets the label of the y-axis.
    label : str, optional
        Sets the label for the plot.
    lab_loc : int, optional
        Defines the position of the legend
    ax : pyplot.Axes, optional
        Use the given axes to make the plot, defaults to the current axes.
    grid : boolean, optional
        If not given defaults to the value defined in splotch.Params.
    plot_kw : dict, optional
        Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are Line2D properties.
    **kwargs: Line2D properties, optional
        kwargs are used to specify matplotlib specific properties such as linecolor, linewidth, antialiasing, etc.
        A list of available `Line2D` properties can be found here:
        https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D

    Returns
    -------
    None
    """

    from splotch.base_func import axes_handler, bin_axis, plot_finalizer

    import numpy as np
    from numbers import Number
    import scipy.stats as stats
    from numpy import percentile
    from functools import partial
    import matplotlib.colors as clr
    import matplotlib.pyplot as plt
    from matplotlib.pyplot import gca, errorbar, rcParams
    
    # Set the current axis
    if ax is not None:
        old_axes=axes_handler(ax)
    else:
        ax=gca()
        old_axes=ax
    
    if ylog is None:
        from splotch.defaults import Params
        ylog = Params.hist1D_yaxis_log
    if bins is None:
        bins = int((len(x))**0.4)

    if not isinstance(stat_y, str):
        try:
            iter(stat_y)
        except TypeError:
            stat_y = [stat_y, stat_y]
    else:
        stat_y = [stat_y, stat_y]

    # Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
    # plot_par = {**plot_kw, **kwargs}  # For Python > 3.5
    plot_par = plot_kw.copy()
    plot_par.update(kwargs)
    if 'linewidth' not in plot_par.keys():
        plot_par['linewidth'] = 0
    if 'elinewidth' not in plot_par.keys():
        plot_par['elinewidth'] = rcParams['lines.linewidth']

    bar_multi = np.ones(2)
    for i in range(len(stat_y)):
        if isinstance(stat_y[i], Number):
            stat_y[i] = partial(percentile, q=stat_y[i])
        elif 'std' in stat_y[i] and len(stat_y[i].replace('std', '')) > 0:
            bar_multi[i] = float(stat_y[i].replace('std', ''))
            stat_y[i] = 'std'

    if isinstance(stat_cen, Number):
        stat_cen = partial(percentile, q=stat_cen)

    temp_x, bins_hist, bins_plot = bin_axis(x, bin_type, bins, log=xlog)
    temp_y = stats.binned_statistic(temp_x, y, statistic=stat_cen, bins=bins_hist)[0]
    if stat_y[0] == stat_y[1]:
        bar_low, bar_high = [stats.binned_statistic(temp_x, y, statistic=stat_y[0], bins=bins_hist)[0]] * 2
    else:
        bar_low = stats.binned_statistic(temp_x, y, statistic=stat_y[0], bins=bins_hist)[0]
        bar_high = stats.binned_statistic(temp_x, y, statistic=stat_y[1], bins=bins_hist)[0]
    if stat_y[0] == 'std':
        bar_low = bar_multi[0] * bar_low
    else:
        bar_low = temp_y - bar_low
    if stat_y[0] == 'std':
        bar_high = bar_multi[0] * bar_high
    else:
        bar_high -= temp_y
    if ylog:
        temp_y = np.where(temp_y == 0, np.nan, temp_y)

    x = stats.binned_statistic(temp_x, x, statistic=stat_cen, bins=bins_hist)[0]
    y = temp_y
    if bar_x:
        bar_x = [x - bins_plot[:-1], bins_plot[1:] - x]

    if bar_x:
        errorbar(x, y, xerr=bar_x, yerr=[bar_low, bar_high], **plot_par)
    else:
        errorbar(x, y, yerr=[bar_low, bar_high], **plot_par)

    if label is not None:
        plt.legend(loc=lab_loc)

    plot_finalizer(xlog, ylog, xlim, ylim, title, xlabel, ylabel, xinvert, yinvert, grid)
    if old_axes is not ax:
        old_axes = axes_handler(old_axes)



def use_style(path):
    """ Loads in splotch configuration file containing a set of splotch style 
        parameters appended on top of a standard matplotlib .mplstyle configuration
        file.
    
    Parameters
    ----------
    path : string
            The path to the configuration file.
    
    Returns
    -------
    None
    
    """
    
    from .defaults import Params
    from .base_func import val_checker
    from matplotlib.pyplot import style
    
    (path)
    
    ### SPLOTCH specific
    # Get list of available parameters within the splotch.Params class
    validParams=[key for key in Params.__dict__.keys() if key[:1] != '_']
    
    # find first Splotch line
    begin=None
    with open(path, 'r') as file:
        for num, line in enumerate(file.readlines()):
            if ("#### splotch configuration" in line.lower()):
                begin=num
                
    spltFile=open(path, 'r')
    allLines=spltFile.readlines()
    spltLines=allLines[begin:]
    pltLines=allLines[:begin]
    spltFile.close()
    pltDict ={}
    
    for line in spltLines:
        if (":" in line): # This line calls a parameter
            par=line.split(":")[0].rstrip()
            if par[0] == '#':
                continue
            else:
                val=line.split(":")[-1].lstrip().replace("\n","")
                
                # Adjust parameter value for booleans and numbers (i.e. floats/integer)
                if ',' in val:
                    val=[val_checker(v) for v in val.split(',')]
                else:
                    val=val_checker(val)
                
                if (par in validParams): # only edit parameters that exist within Params Class
                    setattr(Params,par,val)
    
    for line in pltLines:
        if (":" in line): # This line calls a parameter
            par=line.split(":")[0].rstrip().lstrip()
            if par[0] == '#':
                continue
            else:
                val=line.split('#')[0].split(":")[-1].lstrip().replace("\n","").rstrip()
                
                # Adjust parameter value for booleans and numbers (i.e. floats/integer)
                if ',' in val and "'" not in val:
                    val=[val_checker(v.rstrip()) for v in val.split(',')]
                else:
                    val=val_checker(val.rstrip())
                
                pltDict[par]=val
    style.use(pltDict)
    
    return(None)

def reset_style():
    """ Resets the Splotch parameters to their defaults.
    
    Parameters
    ----------
    None
    
    Returns
    -------
    None
    
    """
    import os
    import splotch
    
    basedir=os.path.dirname(splotch.__file__)
    splotch.use_style("{0}/styles/default.style".format(basedir))
    
    return(None)

